import subprocess

commandlist = {
    'artistools-comparetogsinetwork': ('artistools.gsinetwork', 'main'),

    'artistools-modeldeposition': ('artistools.deposition', 'main_analytical'),

    'getartisspencerfano': ('artistools.nonthermal.solvespencerfanocmd', 'main'),
    'artistools-spencerfano': ('artistools.nonthermal.solvespencerfanocmd', 'main'),

    'listartistimesteps': ('artistools', 'showtimesteptimes'),
    'artistools-timesteptimes': ('artistools', 'showtimesteptimes'),

    'artistools-make1dslicefrom3dmodel': ('artistools.inputmodel.1dslicefrom3d', 'main'),
    'makeartismodel1dslicefromcone': ('artistools.inputmodel.slice1Dfromconein3dmodel', 'main'),
    'makeartismodelbotyanski2017': ('artistools.inputmodel.botyanski2017', 'main'),
    'makeartismodelfromshen2018': ('artistools.inputmodel.shen2018', 'main'),
    'makeartismodelfromlapuente': ('artistools.inputmodel.lapuente', 'main'),
    'makeartismodelscalevelocity': ('artistools.inputmodel.scalevelocity', 'main'),
    'makeartismodelfullymixed': ('artistools.inputmodel.fullymixed', 'main'),
    'makeartismodelsolar_rprocess': ('artistools.inputmodel.rprocess_solar', 'main'),
    'makeartismodelfromsingletrajectory': ('artistools.inputmodel.rprocess_from_trajectory', 'main'),
    'makeartismodelfromparticlegridmap': ('artistools.inputmodel.modelfromhydro', 'main'),
    'makeartismodel': ('artistools.inputmodel.makeartismodel', 'main'),

    'artistools-maketardismodelfromartis': ('artistools.inputmodel.maketardismodelfromartis', 'main'),

    'artistools-maptogrid': ('artistools.inputmodel.maptogrid', 'main'),

    'plotartisdeposition': ('artistools.deposition', 'main'),
    'artistools-deposition': ('artistools.deposition', 'main'),

    'artistools-describeinputmodel': ('artistools.inputmodel.describeinputmodel', 'main'),

    'plotartisestimators': ('artistools.estimators.plotestimators', 'main'),
    'artistools-estimators': ('artistools.estimators', 'main'),

    'artistools-exportmassfractions': ('artistools.estimators.exportmassfractions', 'main'),

    'plotartislightcurve': ('artistools.lightcurve.plotlightcurve', 'main'),
    'artistools-lightcurve': ('artistools.lightcurve', 'main'),

    'plotartislinefluxes': ('artistools.linefluxes', 'main'),
    'artistools-linefluxes': ('artistools.linefluxes', 'main'),

    'plotartismacroatom': ('artistools.macroatom', 'main'),
    'artistools-macroatom': ('artistools.macroatom', 'main'),

    'plotartisnltepops': ('artistools.nltepops.plotnltepops', 'main'),
    'artistools-nltepops': ('artistools.nltepops', 'main'),

    'plotartisnonthermal': ('artistools.nonthermal', 'main'),
    'artistools-nonthermal': ('artistools.nonthermal', 'main'),

    'plotartisradfield': ('artistools.radfield', 'main'),
    'artistools-radfield': ('artistools.radfield', 'main'),

    'plotartisspectrum': ('artistools.spectra.plotspectra', 'main'),
    'artistools-spectrum': ('artistools.spectra', 'main'),

    'plotartistransitions': ('artistools.transitions', 'main'),
    'artistools-transitions': ('artistools.transitions', 'main'),

    'plotartisinitialcomposition': ('artistools.initial_composition', 'main'),
    'artistools-initialcomposition': ('artistools.initial_composition', 'main'),

    'artistools-writecodecomparisondata': ('artistools.writecomparisondata', 'main'),
}

console_scripts = [f'{command} = {submodulename}:{funcname}'
                   for command, (submodulename, funcname) in commandlist.items()]
console_scripts.append('at = artistools:main')
console_scripts.append('artistools = artistools:main')

completioncommands = []
for command in commandlist.keys():
    result = subprocess.run(['register-python-argcomplete', command], stdout=subprocess.PIPE)
    completioncommands.append(result.stdout.decode('utf-8') + '\n')



#!/usr/bin/env python3
"""Artistools - spectra related functions."""
import math
import multiprocessing
from collections import namedtuple
from functools import lru_cache
from functools import partial
from pathlib import Path
import os

import matplotlib as mpl
import matplotlib.pyplot as plt  # needed to get the color map
import numpy as np
import pandas as pd
from astropy import constants as const
from astropy import units as u
import re

import artistools as at
import artistools.radfield
import artistools.packets

fluxcontributiontuple = namedtuple(
    'fluxcontribution', 'fluxcontrib linelabel array_flambda_emission array_flambda_absorption color')


def timeshift_fluxscale_co56law(scaletoreftime, spectime):
    if scaletoreftime is not None:
        # Co56 decay flux scaling
        assert spectime > 150
        return math.exp(float(spectime) / 113.7) / math.exp(scaletoreftime / 113.7)
    else:
        return 1.


def get_exspec_bins():
    MNUBINS = 1000
    NU_MIN_R = 1e13
    NU_MAX_R = 5e15

    c_ang_s = const.c.to('angstrom/s').value

    dlognu = (math.log(NU_MAX_R) - math.log(NU_MIN_R)) / MNUBINS

    bins_nu_lower = np.array(
        [math.exp(math.log(NU_MIN_R) + (m * (dlognu))) for m in range(MNUBINS)])
    # bins_nu_upper = np.array(
    #     [math.exp(math.log(NU_MIN_R) + ((m + 1) * (dlognu))) for m in range(MNUBINS)])
    bins_nu_upper = bins_nu_lower * math.exp(dlognu)
    bins_nu_centre = 0.5 * (bins_nu_lower + bins_nu_upper)

    array_lambdabinedges = np.append(c_ang_s / np.flip(bins_nu_upper), c_ang_s / bins_nu_lower[0])
    array_lambda = c_ang_s / np.flip(bins_nu_centre)
    delta_lambda = np.flip(c_ang_s / bins_nu_lower - c_ang_s / bins_nu_upper)

    return array_lambdabinedges, array_lambda, delta_lambda


def stackspectra(spectra_and_factors):
    """Add spectra using weighting factors, i.e., specout[nu] = spec1[nu] * factor1 + spec2[nu] * factor2 + ...
    spectra_and_factors should be a list of tuples: spectra[], factor"""

    factor_sum = sum([factor for _, factor in spectra_and_factors])

    stackedspectrum = np.zeros_like(spectra_and_factors[0][0], dtype=float)
    for spectrum, factor in spectra_and_factors:
        stackedspectrum += spectrum * factor / factor_sum

    return stackedspectrum


@lru_cache(maxsize=16)
def get_specdata(modelpath, stokesparam=None):
    polarisationdata = False
    if Path(modelpath, 'specpol.out').is_file():
        specfilename = Path(modelpath) / "specpol.out"
        polarisationdata = True
    elif Path(modelpath).is_dir():
        specfilename = at.firstexisting(['spec.out.xz', 'spec.out.gz', 'spec.out'], path=modelpath)
    else:
        specfilename = modelpath

    if polarisationdata:
        # angle = args.plotviewingangle[0]
        stokes_params = get_specpol_data(angle=None, modelpath=modelpath)
        if stokesparam is not None:
            specdata = stokes_params[stokesparam]
        else:
            specdata = stokes_params['I']
    else:
        assert stokesparam is None
        print(f"Reading {specfilename}")
        specdata = pd.read_csv(specfilename, delim_whitespace=True)
        specdata = specdata.rename(columns={'0': 'nu'})

    return specdata


def get_spectrum(
        modelpath, timestepmin: int, timestepmax=-1, fnufilterfunc=None,
        modelnumber=None):
    """Return a pandas DataFrame containing an ARTIS emergent spectrum."""
    if timestepmax < 0:
        timestepmax = timestepmin

    specdata = get_specdata(modelpath)

    nu = specdata.loc[:, 'nu'].values
    arr_tdelta = at.get_timestep_times_float(modelpath, loc='delta')

    f_nu = stackspectra([
        (specdata[specdata.columns[timestep + 1]],
         arr_tdelta[timestep])
        for timestep in range(timestepmin, timestepmax + 1)])

    # best to use the filter on this list because it
    # has regular sampling
    if fnufilterfunc:
        print("Applying filter to ARTIS spectrum")
        f_nu = fnufilterfunc(f_nu)

    dfspectrum = pd.DataFrame({'nu': nu, 'f_nu': f_nu})
    dfspectrum.sort_values(by='nu', ascending=False, inplace=True)

    dfspectrum.eval('lambda_angstroms = @c / nu', local_dict={'c': const.c.to('angstrom/s').value}, inplace=True)
    dfspectrum.eval('f_lambda = f_nu * nu / lambda_angstroms', inplace=True)

    # if 'redshifttoz' in args and args.redshifttoz[modelnumber] != 0:
    # #     plt.plot(dfspectrum['lambda_angstroms'], dfspectrum['f_lambda'], color='k')
    #     z = args.redshifttoz[modelnumber]
    #     dfspectrum['lambda_angstroms'] *= (1 + z)
    # #     plt.plot(dfspectrum['lambda_angstroms'], dfspectrum['f_lambda'], color='r')
    # #     plt.show()
    # #     quit()

    return dfspectrum


def get_spectrum_at_time(modelpath, timestep, time, args, angle=None, res_specdata=None, modelnumber=None):
    if angle is not None:
        if args.plotvspecpol and os.path.isfile(modelpath/'vpkt.txt'):
            spectrum = get_vspecpol_spectrum(modelpath, time, angle, args)
        elif os.path.isfile(modelpath/'specpol_res.out'):
            spectrum = get_res_spectrum(modelpath, timestep, timestep, angle=angle, res_specdata=res_specdata)
        else:
            spectrum = get_spectrum(modelpath, timestep, timestep, modelnumber=modelnumber)
    else:
        spectrum = get_spectrum(modelpath, timestep, timestep, modelnumber=modelnumber)

    return spectrum


def get_spectrum_from_packets_worker(querystr, qlocals, array_lambda, array_lambdabinedges, packetsfile,
                                     use_comovingframe=False, getpacketcount=False, betafactor=None):
    dfpackets = at.packets.readfile(packetsfile, type='TYPE_ESCAPE', escape_type='TYPE_RPKT').query(
        querystr, inplace=False, local_dict=qlocals)

    print(f"  {packetsfile}: {len(dfpackets)} escaped r-packets matching frequency and arrival time ranges ")

    dfpackets.eval('lambda_rf = @c_ang_s / nu_rf', inplace=True, local_dict=qlocals)
    wl_bins = pd.cut(
        x=dfpackets['lambda_rf'], bins=array_lambdabinedges, right=True,
        labels=range(len(array_lambda)), include_lowest=True)

    if use_comovingframe:
        array_energysum_onefile = dfpackets.e_cmf.groupby(wl_bins).sum().values / betafactor
    else:
        array_energysum_onefile = dfpackets.e_rf.groupby(wl_bins).sum().values

    if getpacketcount:
        array_pktcount_onefile = dfpackets.lambda_rf.groupby(wl_bins).count().values
    else:
        array_pktcount_onefile = None

    return array_energysum_onefile, array_pktcount_onefile


def get_spectrum_from_packets(
        modelpath, timelowdays, timehighdays, lambda_min, lambda_max,
        delta_lambda=None, use_comovingframe=None, maxpacketfiles=None, useinternalpackets=False,
        getpacketcount=False):
    """Get a spectrum dataframe using the packets files as input."""
    assert(not useinternalpackets)
    packetsfiles = at.packets.get_packetsfilepaths(modelpath, maxpacketfiles)

    if use_comovingframe:
        modeldata, _, _ = at.inputmodel.get_modeldata(Path(packetsfiles[0]).parent)
        vmax = modeldata.iloc[-1].velocity_outer * u.km / u.s
        betafactor = math.sqrt(1 - (vmax / const.c).decompose().value ** 2)
    else:
        betafactor = None

    c_cgs = const.c.to('cm/s').value
    c_ang_s = const.c.to('angstrom/s').value
    nu_min = c_ang_s / lambda_max
    nu_max = c_ang_s / lambda_min

    if delta_lambda:
        array_lambdabinedges = np.arange(lambda_min, lambda_max + delta_lambda, delta_lambda)
        array_lambda = 0.5 * (array_lambdabinedges[:-1] + array_lambdabinedges[1:])  # bin centres
    else:
        array_lambdabinedges, array_lambda, delta_lambda = get_exspec_bins()

    array_energysum = np.zeros_like(array_lambda, dtype=float)  # total packet energy sum of each bin
    if getpacketcount:
        array_pktcount = np.zeros_like(array_lambda, dtype=int)  # number of packets in each bin

    timelow = timelowdays * u.day.to('s')
    timehigh = timehighdays * u.day.to('s')

    nprocs_read = len(packetsfiles)
    querystr = '@nu_min <= nu_rf < @nu_max and trueemissiontype >= 0 and '
    if not use_comovingframe:
        querystr += '@timelow < (escape_time - (posx * dirx + posy * diry + posz * dirz) / @c_cgs) < @timehigh'
    else:
        querystr += '@timelow < (escape_time * @betafactor) < @timehigh'

    processfile = partial(
        get_spectrum_from_packets_worker, querystr,
        dict(nu_min=nu_min, nu_max=nu_max, timelow=timelow, timehigh=timehigh,
             betafactor=betafactor, c_cgs=c_cgs, c_ang_s=c_ang_s),
        array_lambda, array_lambdabinedges, use_comovingframe=use_comovingframe, getpacketcount=getpacketcount,
        betafactor=betafactor)
    if at.num_processes > 1:
        with multiprocessing.Pool(processes=at.num_processes) as pool:
            results = pool.map(processfile, packetsfiles)
            pool.close()
            pool.join()
            pool.terminate()
    else:
        results = [processfile(p) for p in packetsfiles]

    array_energysum = np.ufunc.reduce(np.add, [r[0] for r in results])
    if getpacketcount:
        array_pktcount += np.ufunc.reduce(np.add, [r[1] for r in results])

    array_flambda = (array_energysum / delta_lambda / (timehigh - timelow) /
                     4 / math.pi / (u.megaparsec.to('cm') ** 2) / nprocs_read)

    dfdict = {
        'lambda_angstroms': array_lambda,
        'f_lambda': array_flambda,
        'energy_sum': array_energysum,
    }

    if getpacketcount:
        dfdict['packetcount'] = array_pktcount

    return pd.DataFrame(dfdict)


@at.diskcache(savezipped=True)
def read_specpol_res(modelpath):
    """Return specpol_res data for a given angle"""
    if Path(modelpath, 'specpol_res.out').is_file():
        specfilename = Path(modelpath) / "specpol_res.out"
    else:
        specfilename = modelpath

    print(f"Reading {specfilename} (in read_specpol_res)")
    specdata = pd.read_csv(specfilename, delim_whitespace=True, header=None, dtype=str)

    res_specdata = at.gather_res_data(specdata)

    # index_to_split = specdata.index[specdata.iloc[:, 1] == specdata.iloc[0, 1]]
    # # print(len(index_to_split))
    # res_specdata = []
    # for i, index_value in enumerate(index_to_split):
    #     if index_value != index_to_split[-1]:
    #         chunk = specdata.iloc[index_to_split[i]:index_to_split[i + 1], :]
    #     else:
    #         chunk = specdata.iloc[index_to_split[i]:, :]
    #     res_specdata.append(chunk)
    # print(res_specdata[0])

    columns = res_specdata[0].iloc[0]
    # print(columns)
    for i, res_spec in enumerate(res_specdata):
        res_specdata[i] = res_specdata[i].rename(columns=columns).drop(res_specdata[i].index[0])
        # These lines remove the Q and U values from the dataframe (I think)
        numberofIvalues = len(res_specdata[i].columns.drop_duplicates())
        res_specdata[i] = res_specdata[i].iloc[:, : numberofIvalues]
        res_specdata[i] = res_specdata[i].astype(float)
        res_specdata[i] = res_specdata[i].to_numpy()

    for i, res_spec in enumerate(res_specdata):
        res_specdata[i] = pd.DataFrame(data=res_specdata[i], columns=columns[:numberofIvalues])
        res_specdata[i] = res_specdata[i].rename(columns={'0': 'nu'})

    return res_specdata


def average_angle_bins(res_specdata, angle, args):
    # Averages over 10 bins to reduce noise

    if args and args.average_every_tenth_viewing_angle:
        for start_bin in np.arange(start=0, stop=100, step=10):
            # print(start_bin)
            for bin_number in range(start_bin+1, start_bin+10):
                # print(bin_number)
                res_specdata[start_bin] += res_specdata[bin_number]
            res_specdata[start_bin] /= 10  # every 10th bin is the average of 10 bins
            print(f'bin number {start_bin} = the average of bins {start_bin} to {start_bin + 9}')

        if angle and angle % 10 == 0:
            print(f"Bin number {angle} is the average of 10 angle bins")
    return res_specdata


def get_res_spectrum(
        modelpath, timestepmin: int, timestepmax=-1, angle=None, res_specdata=None, fnufilterfunc=None,
        args=None):

    """Return a pandas DataFrame containing an ARTIS emergent spectrum."""
    if timestepmax < 0:
        timestepmax = timestepmin

    # print(f"Reading spectrum at timestep {timestepmin}")

    if angle is None:
        angle = args.plotviewingangle[0]

    if res_specdata is None:
        print("Reading specpol_res.out")
        res_specdata = read_specpol_res(modelpath)
        if args and 'average_every_tenth_viewing_angle' in args:
            at.spectra.average_angle_bins(res_specdata, angle, args)

    nu = res_specdata[angle].loc[:, 'nu'].values
    arr_tmid = at.get_timestep_times_float(modelpath, loc='mid')
    arr_tdelta = at.get_timestep_times_float(modelpath, loc='delta')

    # for angle in args.plotviewingangle:
    f_nu = stackspectra([(res_specdata[angle][res_specdata[angle].columns[timestep + 1]],
                          arr_tdelta[timestep])
                         for timestep in range(timestepmin, timestepmax + 1)])

    # best to use the filter on this list because it
    # has regular sampling
    if fnufilterfunc:
        print("Applying filter to ARTIS spectrum")
        f_nu = fnufilterfunc(f_nu)

    dfspectrum = pd.DataFrame({'nu': nu, 'f_nu': f_nu})
    dfspectrum.sort_values(by='nu', ascending=False, inplace=True)

    dfspectrum.eval('lambda_angstroms = @c / nu', local_dict={'c': const.c.to('angstrom/s').value}, inplace=True)
    dfspectrum.eval('f_lambda = f_nu * nu / lambda_angstroms', inplace=True)
    return dfspectrum


def make_virtual_spectra_summed_file(modelpath):
    nprocs = at.get_nprocs(modelpath)
    print("nprocs", nprocs)
    vspecpol_data_old = []  # virtual packet spectra for each observer (all directions and opacity choices)
    vpktconfig = at.get_vpkt_config(modelpath)
    nvirtual_spectra = vpktconfig['nobsdirections'] * vpktconfig['nspectraperobs']
    print(f"nobsdirections {vpktconfig['nobsdirections']} nspectraperobs {vpktconfig['nspectraperobs']} (total observers: {nvirtual_spectra})")
    for mpirank in range(nprocs):
        vspecpolfilename = f'vspecpol_{mpirank}-0.out'
        print(f"Reading rank {mpirank} filename {vspecpolfilename}")
        vspecpolpath = Path(modelpath, vspecpolfilename)
        if not vspecpolpath.is_file():
            vspecpolpath = Path(modelpath, vspecpolfilename + '.gz')
            if not vspecpolpath.is_file():
                print(f'Warning: Could not find {vspecpolpath.relative_to(modelpath.parent)}')
                continue

        vspecpolfile = pd.read_csv(vspecpolpath, delim_whitespace=True, header=None)
        # Where times of timesteps are written out a new virtual spectrum starts
        # Find where the time in row 0, column 1 repeats in any column 1
        index_of_new_spectrum = vspecpolfile.index[vspecpolfile.iloc[:, 1] == vspecpolfile.iloc[0, 1]]
        vspecpol_data = []  # list of all predefined vspectra
        for i, index_spectrum_starts in enumerate(index_of_new_spectrum[:nvirtual_spectra]):
            # todo: this is different to at.gather_res_data() -- could be made to be same format to not repeat code
            if index_spectrum_starts != index_of_new_spectrum[-1]:
                chunk = vspecpolfile.iloc[index_spectrum_starts:index_of_new_spectrum[i+1], :]
            else:
                chunk = vspecpolfile.iloc[index_spectrum_starts:, :]
            vspecpol_data.append(chunk)

        if len(vspecpol_data_old) > 0:
            for i, _ in enumerate(vspecpol_data):
                dftmp = vspecpol_data[i].copy()  # copy of vspectrum number i in a file
                # add copy to the same spectrum number from previous file
                # (don't need to copy row 1 = time or column 1 = freq)
                dftmp.iloc[1:, 1:] += vspecpol_data_old[i].iloc[1:, 1:]
                # spectrum i then equals the sum of all previous files spectrum number i
                vspecpol_data[i] = dftmp
        # update array containing sum of previous files
        vspecpol_data_old = vspecpol_data

    for spec_index, vspecpol in enumerate(vspecpol_data):
        outfile = modelpath / f'vspecpol_total-{spec_index}.out'
        print(f'Saved {outfile}')
        vspecpol.to_csv(outfile, sep=' ', index=False, header=False)


def make_averaged_vspecfiles(args):
    filenames = []
    for vspecfile in os.listdir(args.modelpath[0]):
        if vspecfile.startswith('vspecpol_total-'):
            filenames.append(vspecfile)

    def sorted_by_number(l):

        def convert(text):
            return int(text) if text.isdigit() else text

        def alphanum_key(key):
            return [convert(c) for c in re.split('([0-9]+)', key)]

        return sorted(l, key=alphanum_key)

    filenames = sorted_by_number(filenames)

    for spec_index, filename in enumerate(filenames):  # vspecpol-total files
        vspecdata = []
        for modelpath in args.modelpath:
            vspecdata.append(pd.read_csv(modelpath / filename, delim_whitespace=True, header=None))
        for i in range(1, len(vspecdata)):
            vspecdata[0].iloc[1:, 1:] += vspecdata[i].iloc[1:, 1:]

        vspecdata[0].iloc[1:, 1:] = vspecdata[0].iloc[1:, 1:]/len(vspecdata)
        vspecdata[0].to_csv(args.modelpath[0] / f'vspecpol_averaged-{spec_index}.out',
                            sep=' ', index=False, header=False)


def get_specpol_data(angle=None, modelpath=None, specdata=None):
    if specdata is None:
        if angle is None:
            specfilename = Path(modelpath, 'specpol.out')
        else:
            # alternatively use f'vspecpol_averaged-{angle}.out' ?
            specfilename = Path(modelpath, f'vspecpol_total-{angle}.out')
            if not specfilename.exists():
                print(f"{specfilename} does not exist. Generating all-rank summed vspec files..")
                make_virtual_spectra_summed_file(modelpath=modelpath)

        print(f"Reading {specfilename}")
        specdata = pd.read_csv(specfilename, delim_whitespace=True)
        specdata = specdata.rename(columns={specdata.keys()[0]: 'nu'})

    cols_to_split = []
    stokes_params = {}
    for i, key in enumerate(specdata.keys()):
        if specdata.keys()[1] in key:
            cols_to_split.append(i)

    stokes_params['I'] = pd.concat([specdata['nu'], specdata.iloc[:, cols_to_split[0]: cols_to_split[1]]], axis=1)
    stokes_params['Q'] = pd.concat([specdata['nu'], specdata.iloc[:, cols_to_split[1]: cols_to_split[2]]], axis=1)
    stokes_params['U'] = pd.concat([specdata['nu'], specdata.iloc[:, cols_to_split[2]:]], axis=1)

    for param in ['Q', 'U']:
        stokes_params[param].columns = stokes_params['I'].keys()
        stokes_params[param + '/I'] = pd.concat([specdata['nu'],
                                                 stokes_params[param].iloc[:, 1:]
                                                 / stokes_params['I'].iloc[:, 1:]], axis=1)

    return stokes_params


def get_vspecpol_spectrum(modelpath, timeavg, angle, args, fnufilterfunc=None):
    stokes_params = get_specpol_data(angle, modelpath=modelpath)
    if 'stokesparam' not in args:
        args.stokesparam = 'I'
    vspecdata = stokes_params[args.stokesparam]

    nu = vspecdata.loc[:, 'nu'].values

    arr_tmid = [float(i) for i in vspecdata.columns.values[1:] if i[-2] != '.']
    arr_tdelta = [l1 - l2 for l1, l2 in zip(arr_tmid[1:], arr_tmid[:-1])] + [arr_tmid[-1] - arr_tmid[-2]]

    def match_closest_time(reftime):
        return str("{}".format(min([float(x) for x in arr_tmid], key=lambda x: abs(x - reftime))))

    # if 'timemin' and 'timemax' in args:
    #     timelower = match_closest_time(args.timemin)  # how timemin, timemax are used changed at some point
    #     timeupper = match_closest_time(args.timemax)  # to average over multiple timesteps needs to fix this
    # else:
    timelower = match_closest_time(timeavg)
    timeupper = match_closest_time(timeavg)
    timestepmin = vspecdata.columns.get_loc(timelower)
    timestepmax = vspecdata.columns.get_loc(timeupper)

    f_nu = stackspectra([
        (vspecdata[vspecdata.columns[timestep + 1]],
         arr_tdelta[timestep])
        for timestep in range(timestepmin-1, timestepmax)])

    # best to use the filter on this list because it
    # has regular sampling
    if fnufilterfunc:
        print("Applying filter to ARTIS spectrum")
        f_nu = fnufilterfunc(f_nu)

    dfspectrum = pd.DataFrame({'nu': nu, 'f_nu': f_nu})
    dfspectrum.sort_values(by='nu', ascending=False, inplace=True)

    dfspectrum.eval('lambda_angstroms = @c / nu', local_dict={'c': const.c.to('angstrom/s').value}, inplace=True)
    dfspectrum.eval('f_lambda = f_nu * nu / lambda_angstroms', inplace=True)

    return dfspectrum


@lru_cache(maxsize=4)
def get_flux_contributions(
        modelpath, filterfunc=None, timestepmin=0, timestepmax=None, getemission=True, getabsorption=True,
        use_lastemissiontype=False):

    arr_tmid = at.get_timestep_times_float(modelpath, loc='mid')
    arr_tdelta = at.get_timestep_times_float(modelpath, loc='delta')
    arraynu = at.get_nu_grid(modelpath)
    arraylambda = const.c.to('angstrom/s').value / arraynu
    if not Path(modelpath, 'compositiondata.txt').is_file():
        print('WARNING: compositiondata.txt not found. Using output*.txt instead')
        elementlist = at.get_composition_data_from_outputfile(modelpath)
    else:
        elementlist = at.get_composition_data(modelpath)
    nelements = len(elementlist)

    if getemission:
        if use_lastemissiontype:
            emissionfilenames = ['emission.out.xz', 'emission.out.gz', 'emission.out', 'emissionpol.out']
        else:
            emissionfilenames = ['emissiontrue.out.xz', 'emissiontrue.out.gz', 'emissiontrue.out']

        emissionfilename = at.firstexisting(emissionfilenames, path=modelpath)

        if 'pol' in str(emissionfilename):
            print("This artis run contains polarisation data")
            # File contains I, Q and U and so times are repeated 3 times
            arr_tmid = np.array(arr_tmid.tolist() * 3)

        try:
            emissionfilesize = Path(emissionfilename).stat().st_siplze / 1024 / 1024
            print(f' Reading {emissionfilename} ({emissionfilesize:.2f} MiB)')
        except AttributeError:
            print(f' Reading {emissionfilename}')
        emissiondata = pd.read_csv(emissionfilename, delim_whitespace=True, header=None)
        maxion_float = (emissiondata.shape[1] - 1) / 2 / nelements  # also known as MIONS in ARTIS sn3d.h
        assert maxion_float.is_integer()
        maxion = int(maxion_float)
        print(f' inferred MAXION = {maxion} from emission file using nlements = {nelements} from compositiondata.txt')

        # check that the row count is product of timesteps and frequency bins found in spec.out
        assert emissiondata.shape[0] == len(arraynu) * len(arr_tmid)

    if getabsorption:
        absorptionfilename = at.firstexisting(['absorption.out.xz', 'absorption.out.gz', 'absorption.out',
                                               'absorptionpol.out'], path=modelpath)
        try:
            absorptionfilesize = Path(absorptionfilename).stat().st_size / 1024 / 1024
            print(f' Reading {absorptionfilename} ({absorptionfilesize:.2f} MiB)')
        except AttributeError:
            print(f' Reading {emissionfilename}')
        absorptiondata = pd.read_csv(absorptionfilename, delim_whitespace=True, header=None)
        absorption_maxion_float = absorptiondata.shape[1] / nelements
        assert absorption_maxion_float.is_integer()
        absorption_maxion = int(absorption_maxion_float)
        if not getemission:
            maxion = absorption_maxion
            print(f' inferred MAXION = {maxion} from absorption file using nlements = {nelements}'
                  'from compositiondata.txt')
        else:
            assert absorption_maxion == maxion
        assert absorptiondata.shape[0] == len(arraynu) * len(arr_tmid)
    else:
        absorptiondata = None

    array_flambda_emission_total = np.zeros_like(arraylambda, dtype=float)
    contribution_list = []
    if filterfunc:
        print("Applying filter to ARTIS spectrum")

    for element in range(nelements):
        nions = elementlist.nions[element]
        # nions = elementlist.iloc[element].uppermost_ionstage - elementlist.iloc[element].lowermost_ionstage + 1
        for ion in range(nions):
            ion_stage = ion + elementlist.lowermost_ionstage[element]
            ionserieslist = [(element * maxion + ion, 'bound-bound'),
                             (nelements * maxion + element * maxion + ion, 'bound-free')]

            if element == ion == 0:
                ionserieslist.append((2 * nelements * maxion, 'free-free'))

            for (selectedcolumn, emissiontype) in ionserieslist:
                # if linelabel.startswith('Fe ') or linelabel.endswith("-free"):
                #     continue
                if getemission:
                    array_fnu_emission = stackspectra(
                        [(emissiondata.iloc[timestep::len(arr_tmid), selectedcolumn].values,
                          arr_tdelta[timestep])
                         for timestep in range(timestepmin, timestepmax + 1)])
                else:
                    array_fnu_emission = np.zeros_like(arraylambda, dtype=float)

                if absorptiondata is not None and selectedcolumn < nelements * maxion:  # bound-bound process
                    array_fnu_absorption = stackspectra(
                        [(absorptiondata.iloc[timestep::len(arr_tmid), selectedcolumn].values,
                          arr_tdelta[timestep])
                         for timestep in range(timestepmin, timestepmax + 1)])
                else:
                    array_fnu_absorption = np.zeros_like(arraylambda, dtype=float)

                # best to use the filter on fnu (because it hopefully has regular sampling)
                if filterfunc:
                    array_fnu_emission = filterfunc(array_fnu_emission)
                    if selectedcolumn <= nelements * maxion:
                        array_fnu_absorption = filterfunc(array_fnu_absorption)

                array_flambda_emission = array_fnu_emission * arraynu / arraylambda
                array_flambda_absorption = array_fnu_absorption * arraynu / arraylambda

                array_flambda_emission_total += array_flambda_emission
                fluxcontribthisseries = (
                    abs(np.trapz(array_fnu_emission, x=arraynu)) + abs(np.trapz(array_fnu_absorption, x=arraynu)))

                if emissiontype == 'bound-bound':
                    linelabel = at.get_ionstring(elementlist.Z[element], ion_stage)
                elif emissiontype != 'free-free':
                    linelabel = f'{at.get_ionstring(elementlist.Z[element], ion_stage)} {emissiontype}'
                else:
                    linelabel = f'{emissiontype}'

                contribution_list.append(
                    fluxcontributiontuple(fluxcontrib=fluxcontribthisseries, linelabel=linelabel,
                                          array_flambda_emission=array_flambda_emission,
                                          array_flambda_absorption=array_flambda_absorption,
                                          color=None))

    return contribution_list, array_flambda_emission_total


@lru_cache(maxsize=4)
def get_flux_contributions_from_packets(
        modelpath, timelowerdays, timeupperdays, lambda_min, lambda_max, delta_lambda=None,
        getemission=True, getabsorption=True, maxpacketfiles=None, filterfunc=None, groupby='ion', modelgridindex=None,
        use_comovingframe=False, use_lastemissiontype=False, useinternalpackets=False, emissionvelocitycut=None):

    assert groupby in [None, 'ion', 'line', 'upperterm', 'terms']

    if groupby in ['terms', 'upperterm']:
        adata = at.atomic.get_levels(modelpath)

    def get_emprocesslabel(emtype):
        if emtype >= 0:
            line = linelist[emtype]
            if groupby == 'line':
                # if line.atomic_number != 26 or line.ionstage != 2:
                #     return 'non-Fe II ions'
                return (f'{at.get_ionstring(line.atomic_number, line.ionstage)} '
                        f'λ{line.lambda_angstroms:.0f} '
                        f'({line.upperlevelindex}-{line.lowerlevelindex})')
            elif groupby == 'terms':
                upper_config = adata.query(
                    'Z == @line.atomic_number and ion_stage == @line.ionstage', inplace=False
                    ).iloc[0].levels.iloc[line.upperlevelindex].levelname
                upper_term_noj = upper_config.split('_')[-1].split('[')[0]
                lower_config = adata.query(
                    'Z == @line.atomic_number and ion_stage == @line.ionstage', inplace=False
                    ).iloc[0].levels.iloc[line.lowerlevelindex].levelname
                lower_term_noj = lower_config.split('_')[-1].split('[')[0]
                return f'{at.get_ionstring(line.atomic_number, line.ionstage)} {upper_term_noj}->{lower_term_noj}'
            elif groupby == 'upperterm':
                upper_config = adata.query(
                    'Z == @line.atomic_number and ion_stage == @line.ionstage', inplace=False
                    ).iloc[0].levels.iloc[line.upperlevelindex].levelname
                upper_term_noj = upper_config.split('_')[-1].split('[')[0]
                return f'{at.get_ionstring(line.atomic_number, line.ionstage)} {upper_term_noj}'
            return f'{at.get_ionstring(line.atomic_number, line.ionstage)} bound-bound'
        elif emtype == -9999999:
            return 'free-free'

        bflist = at.get_bflist(modelpath)
        bfindex = -emtype - 1
        if bfindex in bflist:
            (atomic_number, ionstage, level) = bflist[bfindex][:3]
            if groupby == 'line':
                return f'{at.get_ionstring(atomic_number, ionstage)} bound-free {level}'
            return f'{at.get_ionstring(atomic_number, ionstage)} bound-free'
        return f'? bound-free (bfindex={bfindex})'

    def get_absprocesslabel(abstype):
        if abstype >= 0:
            line = linelist[abstype]
            if groupby == 'line':
                return (f'{at.get_ionstring(line.atomic_number, line.ionstage)} '
                        f'λ{line.lambda_angstroms:.0f} '
                        f'({line.upperlevelindex}-{line.lowerlevelindex})')
            return f'{at.get_ionstring(line.atomic_number, line.ionstage)} bound-bound'
        if abstype == -1:
            return 'free-free'
        if abstype == -2:
            return 'bound-free'
        return '? other absorp.'

    if delta_lambda:
        array_lambdabinedges = np.arange(lambda_min, lambda_max + delta_lambda, delta_lambda)
        array_lambda = 0.5 * (array_lambdabinedges[:-1] + array_lambdabinedges[1:])  # bin centres
    else:
        array_lambdabinedges, array_lambda, delta_lambda = get_exspec_bins()

    if use_comovingframe:
        modeldata, _, _ = at.inputmodel.get_modeldata(modelpath)
        vmax = modeldata.iloc[-1].velocity_outer * u.km / u.s
        betafactor = math.sqrt(1 - (vmax / const.c).decompose().value ** 2)

    import artistools.packets
    packetsfiles = at.packets.get_packetsfilepaths(modelpath, maxpacketfiles)

    linelist = at.get_linelist(modelpath)

    energysum_spectrum_emission_total = np.zeros_like(array_lambda, dtype=float)
    array_energysum_spectra = {}

    timelow = timelowerdays * u.day.to('s')
    timehigh = timeupperdays * u.day.to('s')

    nprocs_read = len(packetsfiles)
    c_cgs = const.c.to('cm/s').value
    c_ang_s = const.c.to('angstrom/s').value
    nu_min = c_ang_s / lambda_max
    nu_max = c_ang_s / lambda_min

    if useinternalpackets:
        emtypecolumn = 'emissiontype'
    else:
        emtypecolumn = 'emissiontype' if use_lastemissiontype else 'trueemissiontype'

    for index, packetsfile in enumerate(packetsfiles):
        if useinternalpackets:
            # if we're using packets*.out files, these packets are from the last timestep
            t_seconds = at.get_timestep_times_float(modelpath, loc='start')[-1] * u.day.to('s')

            if modelgridindex is not None:
                v_inner = at.inputmodel.get_modeldata(modelpath)[0]['velocity_inner'].iloc[modelgridindex] * 1e5
                v_outer = at.inputmodel.get_modeldata(modelpath)[0]['velocity_outer'].iloc[modelgridindex] * 1e5
            else:
                v_inner = 0.
                v_outer = at.inputmodel.get_modeldata(modelpath)[0]['velocity_outer'].iloc[-1] * 1e5

            r_inner = t_seconds * v_inner
            r_outer = t_seconds * v_outer

            dfpackets = at.packets.readfile(packetsfile, type='TYPE_RPKT')
            print("Using non-escaped internal r-packets")
            dfpackets.query(f'type_id == {at.packets.type_ids["TYPE_RPKT"]} and @nu_min <= nu_rf < @nu_max',
                            inplace=True)
            if modelgridindex is not None:
                assoc_cells, mgi_of_propcells = at.get_grid_mapping(modelpath=modelpath)
                # dfpackets.eval(f'velocity = sqrt(posx ** 2 + posy ** 2 + posz ** 2) / @t_seconds', inplace=True)
                # dfpackets.query(f'@v_inner <= velocity <= @v_outer',
                #                 inplace=True)
                dfpackets.query('where in @assoc_cells[@modelgridindex]', inplace=True)
            print(f"  {len(dfpackets)} internal r-packets matching frequency range")
        else:
            dfpackets = at.packets.readfile(packetsfile, type='TYPE_ESCAPE', escape_type='TYPE_RPKT')
            dfpackets.query(
                '@nu_min <= nu_rf < @nu_max and ' +
                ('@timelow < (escape_time - (posx * dirx + posy * diry + posz * dirz) / @c_cgs) < @timehigh'
                 if not use_comovingframe else
                 '@timelow < escape_time * @betafactor < @timehigh'),
                inplace=True)
            print(f"  {len(dfpackets)} escaped r-packets matching frequency and arrival time ranges")

            if emissionvelocitycut:
                dfpackets = at.packets.add_derived_columns(
                    dfpackets, modelpath, ['emission_velocity'])

                dfpackets.query('(emission_velocity / 1e5) > @emissionvelocitycut', inplace=True)

        if np.isscalar(delta_lambda):
            dfpackets.eval('xindex = floor((@c_ang_s / nu_rf - @lambda_min) / @delta_lambda)', inplace=True)
            if getabsorption:
                dfpackets.eval('xindexabsorbed = floor((@c_ang_s / absorption_freq - @lambda_min) / @delta_lambda)',
                               inplace=True)
        else:
            dfpackets['xindex'] = np.digitize(c_ang_s / dfpackets.nu_rf, bins=array_lambdabinedges, right=True) - 1
            if getabsorption:
                dfpackets['xindexabsorbed'] = np.digitize(
                    c_ang_s / dfpackets.absorption_freq, bins=array_lambdabinedges, right=True) - 1

        for _, packet in dfpackets.iterrows():
            lambda_rf = c_ang_s / packet.nu_rf
            xindex = int(packet.xindex)
            assert xindex >= 0

            pkt_en = packet.e_cmf / betafactor if use_comovingframe else packet.e_rf

            energysum_spectrum_emission_total[xindex] += pkt_en

            if getemission:
                # if emtype >= 0 and linelist[emtype].upperlevelindex <= 80:
                #     continue
                # emprocesskey = get_emprocesslabel(packet.emissiontype)
                emprocesskey = get_emprocesslabel(packet[emtypecolumn])
                # print('packet lambda_cmf: {c_ang_s / packet.nu_cmf}.1f}, lambda_rf {lambda_rf:.1f}, {emprocesskey}')

                if emprocesskey not in array_energysum_spectra:
                    array_energysum_spectra[emprocesskey] = (
                        np.zeros_like(array_lambda, dtype=float), np.zeros_like(array_lambda, dtype=float))

                array_energysum_spectra[emprocesskey][0][xindex] += pkt_en

            if getabsorption:
                abstype = packet.absorption_type
                if abstype > 0:
                    absprocesskey = get_absprocesslabel(abstype)

                    xindexabsorbed = int(packet.xindexabsorbed)  # bin by absorption wavelength
                    # xindexabsorbed = xindex  # bin by final escaped wavelength

                    if absprocesskey not in array_energysum_spectra:
                        array_energysum_spectra[absprocesskey] = (
                            np.zeros_like(array_lambda, dtype=float), np.zeros_like(array_lambda, dtype=float))

                    array_energysum_spectra[absprocesskey][1][xindexabsorbed] += pkt_en

    if useinternalpackets:
        volume = 4 / 3. * math.pi * (r_outer ** 3 - r_inner ** 3)
        if modelgridindex:
            volume_shells = volume
            assoc_cells, mgi_of_propcells = at.get_grid_mapping(modelpath=modelpath)
            volume = (at.get_wid_init_at_tmin(modelpath) * t_seconds / (
                at.get_inputparams(modelpath)['tmin'] * u.day.to('s'))) ** 3 * len(assoc_cells[modelgridindex])
            print('volume', volume, 'shell volume', volume_shells, '-------------------------------------------------')
        normfactor = c_cgs / 4 / math.pi / delta_lambda / volume / nprocs_read
    else:
        normfactor = (1. / delta_lambda / (timehigh - timelow) / 4 / math.pi
                      / (u.megaparsec.to('cm') ** 2) / nprocs_read)

    array_flambda_emission_total = energysum_spectrum_emission_total * normfactor

    contribution_list = []
    for (groupname,
         (energysum_spec_emission, energysum_spec_absorption)) in array_energysum_spectra.items():
        array_flambda_emission = energysum_spec_emission * normfactor

        array_flambda_absorption = energysum_spec_absorption * normfactor

        fluxcontribthisseries = (
            abs(np.trapz(array_flambda_emission, x=array_lambda)) +
            abs(np.trapz(array_flambda_absorption, x=array_lambda)))

        linelabel = groupname.replace(' bound-bound', '')

        contribution_list.append(
            fluxcontributiontuple(fluxcontrib=fluxcontribthisseries, linelabel=linelabel,
                                  array_flambda_emission=array_flambda_emission,
                                  array_flambda_absorption=array_flambda_absorption,
                                  color=None))

    return contribution_list, array_flambda_emission_total, array_lambda


def sort_and_reduce_flux_contribution_list(
        contribution_list_in, maxseriescount, arraylambda_angstroms,
        fixedionlist=None, hideother=False, greyscale=False):

    if fixedionlist:
        # sort in manual order
        def sortkey(x):
            return (fixedionlist.index(x.linelabel) if x.linelabel in fixedionlist
                    else len(fixedionlist) + 1, -x.fluxcontrib)
    else:
        # sort descending by flux contribution
        def sortkey(x): return -x.fluxcontrib

    contribution_list = sorted(contribution_list_in, key=sortkey)

    # combine the items past maxseriescount or not in manual list into a single item
    remainder_flambda_emission = np.zeros_like(arraylambda_angstroms, dtype=float)
    remainder_flambda_absorption = np.zeros_like(arraylambda_angstroms, dtype=float)
    remainder_fluxcontrib = 0

    if greyscale:
        hatches = artistools.spectra.plot.hatches
        seriescount = len(fixedionlist) if fixedionlist else maxseriescount
        colorcount = math.ceil(seriescount / 1. / len(hatches))
        greylist = [str(x) for x in np.linspace(0.4, 0.9, colorcount, endpoint=True)]
        color_list = []
        for c in range(colorcount):
            for h in hatches:
                color_list.append(greylist[c])
        # color_list = list(plt.get_cmap('tab20')(np.linspace(0, 1.0, 20)))
        mpl.rcParams['hatch.linewidth'] = 0.1
        # TODO: remove???
        color_list = list(plt.get_cmap('tab20')(np.linspace(0, 1.0, 20)))
    else:
        color_list = list(plt.get_cmap('tab20')(np.linspace(0, 1.0, 20)))

    contribution_list_out = []
    numotherprinted = 0
    maxnumotherprinted = 20
    entered_other = False
    plotted_ion_list = []
    for index, row in enumerate(contribution_list):
        if fixedionlist and row.linelabel in fixedionlist:
            contribution_list_out.append(row._replace(color=color_list[fixedionlist.index(row.linelabel)]))
        elif not fixedionlist and index < maxseriescount:
            contribution_list_out.append(row._replace(color=color_list[index]))
            plotted_ion_list.append(row.linelabel)
        else:
            remainder_fluxcontrib += row.fluxcontrib
            remainder_flambda_emission += row.array_flambda_emission
            remainder_flambda_absorption += row.array_flambda_absorption
            if not entered_other:
                print(f"  Other (top {maxnumotherprinted}):")
                entered_other = True

        if numotherprinted < maxnumotherprinted:
            integemiss = abs(np.trapz(row.array_flambda_emission, x=arraylambda_angstroms))
            integabsorp = abs(np.trapz(-row.array_flambda_absorption, x=arraylambda_angstroms))
            if integabsorp > 0. and integemiss > 0.:
                print(f'{row.fluxcontrib:.1e}, emission {integemiss:.1e}, '
                      f"absorption {integabsorp:.1e} [erg/s/cm^2]: '{row.linelabel}'")
            elif integemiss > 0.:
                print(f"  emission {integemiss:.1e} [erg/s/cm^2]: '{row.linelabel}'")
            else:
                print(f"absorption {integabsorp:.1e} [erg/s/cm^2]: '{row.linelabel}'")

            if entered_other:
                numotherprinted += 1

    if not fixedionlist:
        cmdarg = "'" + "' '".join(plotted_ion_list) + "'"
        print('To reuse this ion/process contribution list, pass the following command-line argument: ')
        print(f'     -fixedionlist {cmdarg}')

    if remainder_fluxcontrib > 0. and not hideother:
        contribution_list_out.append(fluxcontributiontuple(
            fluxcontrib=remainder_fluxcontrib, linelabel='Other',
            array_flambda_emission=remainder_flambda_emission, array_flambda_absorption=remainder_flambda_absorption,
            color='grey'))

    return contribution_list_out


def print_integrated_flux(arr_f_lambda, arr_lambda_angstroms, distance_megaparsec=1.):
    integrated_flux = abs(np.trapz(arr_f_lambda, x=arr_lambda_angstroms)) * u.erg / u.s / (u.cm ** 2)
    print(f' integrated flux ({arr_lambda_angstroms.min():.1f} to '
          f'{arr_lambda_angstroms.max():.1f} A): {integrated_flux:.3e}')
    # luminosity = integrated_flux * 4 * math.pi * (distance_megaparsec * u.megaparsec ** 2)
    # print(f'(L={luminosity.to("Lsun"):.3e})')


def get_line_flux(lambda_low, lambda_high, arr_f_lambda, arr_lambda_angstroms):
    index_low, index_high = [np.searchsorted(arr_lambda_angstroms, wl, side="left") for wl in (lambda_low, lambda_high)]
    flux_integral = abs(np.trapz(arr_f_lambda[index_low:index_high], x=arr_lambda_angstroms[index_low:index_high]))
    return flux_integral


def print_floers_line_ratio(modelpath, timedays, arr_f_lambda, arr_lambda_angstroms):
    f_12570 = get_line_flux(12570 - 200, 12570 + 200, arr_f_lambda, arr_lambda_angstroms)
    f_7155 = get_line_flux(7000, 7350, arr_f_lambda, arr_lambda_angstroms)
    print(f'f_12570 {f_12570:.2e} f_7155 {f_7155:.2e}')
    if f_7155 > 0 and f_12570 > 0:
        fratio = f_12570 / f_7155
        print(f'f_12570/f_7122 = {fratio:.2e} (log10 is {math.log10(fratio):.2e})')
        outfilename = f"fe2_nir_vis_ratio_{os.path.basename(modelpath)}.txt"
        print(f' saved to {outfilename}')
        with open(outfilename, "a+") as f:
            f.write(f'{timedays:.1f} {fratio:.3e}\n')


def get_reference_spectrum(filename):

    if Path(filename).is_file():
        filepath = Path(filename)
    else:
        filepath = Path(at.PYDIR, 'data', 'refspectra', filename)

    metadata = at.get_file_metadata(filepath)

    flambdaindex = metadata.get('f_lambda_columnindex', 1)

    specdata = pd.read_csv(filepath, delim_whitespace=True, header=None, comment='#',
                           names=['lambda_angstroms', 'f_lambda'], usecols=[0, flambdaindex])

    # new_lambda_angstroms = []
    # binned_flux = []
    #
    # wavelengths = specdata['lambda_angstroms']
    # fluxes = specdata['f_lambda']
    # nbins = 10
    #
    # for i in np.arange(start=0, stop=len(wavelengths) - nbins, step=nbins):
    #     new_lambda_angstroms.append(wavelengths[i + int(nbins / 2)])
    #     sum_flux = 0
    #     for j in range(i, i + nbins):
    #
    #         if not math.isnan(fluxes[j]):
    #             print(fluxes[j])
    #             sum_flux += fluxes[j]
    #     binned_flux.append(sum_flux / nbins)
    #
    # filtered_specdata = pd.DataFrame(new_lambda_angstroms, columns=['lambda_angstroms'])
    # filtered_specdata['f_lamba'] = binned_flux
    # print(filtered_specdata)
    # plt.plot(specdata['lambda_angstroms'], specdata['f_lambda'])
    # plt.plot(new_lambda_angstroms, binned_flux)
    #
    # filtered_specdata.to_csv('/Users/ccollins/artis_nebular/artistools/artistools/data/refspectra/' + name,
    #                          index=False, header=False, sep=' ')

    if 'a_v' in metadata or 'e_bminusv' in metadata:
        print('Correcting for reddening')
        from extinction import apply, ccm89

        specdata['f_lambda'] = apply(
            ccm89(specdata['lambda_angstroms'].values, a_v=-metadata['a_v'], r_v=metadata['r_v'], unit='aa'),
            specdata['f_lambda'].values)

    if 'z' in metadata:
        print('Correcting for redshift')
        specdata['lambda_angstroms'] /= (1 + metadata['z'])

    return specdata, metadata


def write_flambda_spectra(modelpath, args):
    """Write out spectra to text files.

    Writes lambda_angstroms and f_lambda to .txt files for all timesteps and create
    a text file containing the time in days for each timestep.
    """
    outdirectory = Path(modelpath, 'spectra')

    outdirectory.mkdir(parents=True, exist_ok=True)

    if Path(modelpath, 'specpol.out').is_file():
        specfilename = modelpath / 'specpol.out'
        specdata = pd.read_csv(specfilename, delim_whitespace=True)
        timearray = [i for i in specdata.columns.values[1:] if i[-2] != '.']
    else:
        specfilename = at.firstexisting(['spec.out.xz', 'spec.out.gz', 'spec.out'], path=modelpath)
        specdata = pd.read_csv(specfilename, delim_whitespace=True)
        timearray = specdata.columns.values[1:]

    number_of_timesteps = len(timearray)

    if not args.timestep:
        args.timestep = f'0-{number_of_timesteps - 1}'

    (timestepmin, timestepmax, args.timemin, args.timemax) = at.get_time_range(
        modelpath, args.timestep, args.timemin, args.timemax, args.timedays)

    with open(outdirectory / 'spectra_list.txt', 'w+') as spectra_list:

        arr_tmid = at.get_timestep_times_float(modelpath, loc='mid')

        for timestep in range(timestepmin, timestepmax + 1):

            dfspectrum = get_spectrum(modelpath, timestep, timestep)
            tmid = arr_tmid[timestep]

            outfilepath = outdirectory / f'spectrum_ts{timestep:02.0f}_{tmid:.0f}d.txt'

            with open(outfilepath, 'w') as spec_file:

                spec_file.write('#lambda f_lambda_1Mpc\n')
                spec_file.write('#[A] [erg/s/cm2/A]\n')

                dfspectrum.to_csv(spec_file, header=False, sep=' ', index=False,
                                  columns=["lambda_angstroms", "f_lambda"])

            spectra_list.write(str(outfilepath.absolute()) + '\n')

    with open(outdirectory / 'time_list.txt', 'w+') as time_list:
        for time in arr_tmid:
            time_list.write(f'{str(time)} \n')

    print(f'Saved in {outdirectory}')



import numpy as np


def make_downscaled_3d_grid(modelpath, inputgridsize=200, outputgridsize=50):
    """Should be same as downscale_3d_grid.pro
    Expects a 3D model with grid^3 cells and outputs 3D model with smallgrid^3 cells"""

    grid = int(inputgridsize)
    smallgrid = int(outputgridsize)

    merge = grid / smallgrid
    merge = int(merge)

    modelfile = modelpath / "model.txt"
    abundancefile = modelpath / "abundances.txt"
    smallmodelfile = modelpath / "model_small.txt"
    smallabundancefile = modelpath / "abundances_small.txt"

    rho = np.zeros((grid, grid, grid))
    ffe = np.zeros((grid, grid, grid))
    fni = np.zeros((grid, grid, grid))
    fco = np.zeros((grid, grid, grid))
    ffe52 = np.zeros((grid, grid, grid))
    fcr48 = np.zeros((grid, grid, grid))
    abund = np.zeros((grid, grid, grid, 31))
    abread = np.zeros(31)

    print("reading abundance file")
    with open(abundancefile, 'r') as sourceabundancefile:
        for x in range(0, grid):
            for y in range(0, grid):
                for z in range(0, grid):
                    abread = sourceabundancefile.readline().split()
                    abund[x, y, z] = abread

    print("reading model file")
    with open(modelfile, 'r') as sourcemodelfile:
        x = sourcemodelfile.readline()
        t_model = sourcemodelfile.readline()
        vmax = sourcemodelfile.readline()

        for x in range(0, grid):
            for y in range(0, grid):
                for z in range(0, grid):
                    dum1, dum2, dum3, dum4, rhoread = sourcemodelfile.readline().split()
                    rho[x, y, z] = rhoread
                    fferead, fniread, fcoread, ffe52read, fcr48read = sourcemodelfile.readline().split()
                    ffe[x, y, z] = fferead
                    fni[x, y, z] = fniread
                    fco[x, y, z] = fcoread
                    ffe52[x, y, z] = ffe52read
                    fcr48[x, y, z] = fcr48read

    rho_small = np.zeros((smallgrid, smallgrid, smallgrid))
    ffe_small = np.zeros((smallgrid, smallgrid, smallgrid))
    fni_small = np.zeros((smallgrid, smallgrid, smallgrid))
    fco_small = np.zeros((smallgrid, smallgrid, smallgrid))
    ffe52_small = np.zeros((smallgrid, smallgrid, smallgrid))
    fcr48_small = np.zeros((smallgrid, smallgrid, smallgrid))
    abund_small = np.zeros((smallgrid, smallgrid, smallgrid, 31))

    for x in range(0, smallgrid):
        for y in range(0, smallgrid):
            for z in range(0, smallgrid):
                for xx in range(0, merge):
                    for yy in range(0, merge):
                        for zz in range(0, merge):
                            rho_small[x, y, z] += rho[x * merge + xx, y * merge + yy, z * merge + zz]
                            ffe_small[x, y, z] += ffe[x * merge + xx, y * merge + yy, z * merge + zz] * rho[x * merge + xx, y * merge + yy, z * merge + zz]
                            fni_small[x, y, z] += fni[x * merge + xx, y * merge + yy, z * merge + zz] * rho[x * merge + xx, y * merge + yy, z * merge + zz]
                            fco_small[x, y, z] += fco[x * merge + xx, y * merge + yy, z * merge + zz] * rho[x * merge + xx, y * merge + yy, z * merge + zz]
                            ffe52_small[x, y, z] += ffe52[x * merge + xx, y * merge + yy, z * merge + zz] * rho[x * merge + xx, y * merge + yy, z * merge + zz]
                            fcr48_small[x, y, z] += fcr48[x * merge + xx, y * merge + yy, z * merge + zz] * rho[x * merge + xx, y * merge + yy, z * merge + zz]
                            abund_small[x, y, z, :] += abund[x * merge + xx, y * merge + yy, z * merge + zz] * rho[x * merge + xx, y * merge + yy, z * merge + zz]

    for x in range(0, smallgrid):
        for y in range(0, smallgrid):
            for z in range(0, smallgrid):
                if (rho_small[x, y, z] > 0):
                    ffe_small[x, y, z] /= rho_small[x, y, z]
                    fni_small[x, y, z] /= rho_small[x, y, z]
                    fco_small[x, y, z] /= rho_small[x, y, z]
                    ffe52_small[x, y, z] /= rho_small[x, y, z]
                    fcr48_small[x, y, z] /= rho_small[x, y, z]
                    for i in range(1, 31):  # check this
                        abund_small[x, y, z, i] /= rho_small[x, y, z]
                    rho_small[x, y, z] /= merge ** 3

    print("writing abundance file")
    i = 0
    with open(modelpath / smallabundancefile, 'w') as newabundancefile:
        for x in range(0, smallgrid):
            for y in range(0, smallgrid):
                for z in range(0, smallgrid):
                    line = abund_small[x, y, z, :].tolist()  # index 1:30 are abundances
                    line[0] = int(i+1)  # replace index 0 with index id
                    i += 1
                    newabundancefile.writelines("%s " % item for item in line)
                    newabundancefile.writelines("\n")

    print("writing model file")
    xmax = float(vmax)*float(t_model)*3600*24
    i = 0
    with open(modelpath / smallmodelfile, 'w') as newmodelfile:
        gridsize = int(smallgrid**3)
        newmodelfile.write(f'{gridsize}\n')
        newmodelfile.write(f'{t_model}')
        newmodelfile.write(f'{vmax}')

        for x in range(0, smallgrid):
            for y in range(0, smallgrid):
                for z in range(0, smallgrid):
                    line1 = [int(i+1), -xmax + 2 * x * xmax / smallgrid, -xmax + 2 * y * xmax / smallgrid,  -xmax + 2 * z * xmax / smallgrid, rho_small[x, y, z]]
                    line2 = [ffe_small[x, y, z], fni_small[x, y, z], fco_small[x, y, z], ffe52_small[x, y, z], fcr48_small[x, y, z]]
                    i += 1
                    newmodelfile.writelines("%s " % item for item in line1)
                    newmodelfile.writelines("\n")
                    newmodelfile.writelines("%s " % item for item in line2)
                    newmodelfile.writelines("\n")



#!/usr/bin/env python3
# PYTHON_ARGCOMPLETE_OK

import argcomplete
import argparse
from yaml import dump as yamldump

from pathlib import Path
from artistools import CustomArgHelpFormatter, get_model_name, get_atomic_number
from artistools.inputmodel import get_modeldata


def addargs(parser):

    parser.add_argument('-inputpath', '-i', default='.',
                        help='Path of input ARTIS model')

    parser.add_argument('-temperature', '-T', default=10000,
                        help='Temperature to use in TARDIS file')

    parser.add_argument('-dilution_factor', '-W', default=1.,
                        help='Dilution factor to use in TARDIS file')

    parser.add_argument('-abundtype', choices=['nuclear', 'elemental'], default='elemental',
                        help='Output nuclear or elemental abundances')

    parser.add_argument('-maxatomicnumber', default=92,
                        help='Maximum atomic number for elemental abundances')

    parser.add_argument('-outputpath', '-o', default='.',
                        help='Path of output TARDIS model file')


def main(args=None, argsraw=None, **kwargs) -> None:
    if args is None:
        parser = argparse.ArgumentParser(
            formatter_class=CustomArgHelpFormatter,
            description='Convert an ARTIS format model to TARDIS format.')

        addargs(parser)
        parser.set_defaults(**kwargs)
        argcomplete.autocomplete(parser)
        args = parser.parse_args(argsraw)

    temperature = args.temperature
    dilution_factor = args.dilution_factor

    modelpath = Path(args.inputpath)

    dfmodel, t_model_init_days, _ = get_modeldata(
        modelpath, get_abundances=(args.abundtype == 'elemental'), dimensions=1)

    dfmodel.eval('rho = 10 ** logrho', inplace=True)

    if args.abundtype == 'nuclear':
        # nuclide abundances
        listspecies = [
            col[2:] for col in dfmodel.columns
            if col.startswith('X_') and col.upper() != 'X_FEGROUP' and col[-1].isdigit()]
    else:
        # nuclide abundances
        listspecies = [
            col[2:] for col in dfmodel.columns
            if col.startswith('X_') and col.upper() != 'X_FEGROUP' and not col[-1].isdigit()]

    if args.maxatomicnumber and args.maxatomicnumber > 0:
        listspecies = [species for species in listspecies if get_atomic_number(species) <= args.maxatomicnumber]

    modelname = get_model_name(modelpath)
    outputfilepath = Path(args.outputpath, f'{modelname}.csvy')
    dictmeta = {
        'name': modelname,
        'description': 'This model was converted from ARTIS format with artistools',
        'model_density_time_0': f'{t_model_init_days} day',
        'model_isotope_time_0': f'{t_model_init_days} day',
        'tardis_model_config_version': 'v1.0',
        'datatype': {
            'fields': [
                {'name': 'velocity',
                 'unit': 'km/s',
                 'desc': 'velocities of shell outer boundaries'},
                {'name': 'density',
                 'unit': 'g/cm^3',
                 'desc': 'density of shell'},
                {'name': 't_rad',
                 'unit': 'K',
                 'desc': 'radiative temperature'},
                {'name': 'dilution_factor',
                 'desc': 'dilution factor of shell'},
                *[{'name': strnuc, 'desc': f'fractional {strnuc} abundance'} for strnuc in listspecies]
            ]
        }
    }
    with open(outputfilepath, 'w') as fileout:
        fileout.write('---\n')
        yamldump(dictmeta, fileout, sort_keys=False)
        fileout.write('---\n')
        fileout.write(','.join(['velocity', 'density', 't_rad', 'dilution_factor', *listspecies]))
        fileout.write('\n')

        # fileout.write(f'{0.},{0.:.4e},{0.},{0.},{",".join([f"{0.:.4e}" for _ in listspecies])}\n')

        for cell in dfmodel.itertuples(index=False):
            abundlist = [f'{getattr(cell, "X_" + strnuc):.4e}' for strnuc in listspecies]
            fileout.write(
                f'{cell.velocity_outer},{cell.rho:.4e},{temperature},{dilution_factor},{",".join(abundlist)}\n')

    print(f'Saved {outputfilepath}')


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import artistools as at
import artistools.spectra

from pathlib import Path
import numpy as np
import pandas as pd
from astropy import units as u


def get_bol_lc_from_spec(modelpath):
    res_specdata = at.spectra.read_specpol_res(modelpath)
    # print(res_specdata)
    timearray = res_specdata[0].columns.values[1:]
    times = [time for time in timearray if 5 < float(time) < 80]
    lightcurvedata = {'time': times}

    for angle in range(len(res_specdata)):
        bol_luminosity = []
        for timestep, time in enumerate(timearray):
            time = float(time)
            if 5 < time < 80:
                spectrum = at.spectra.get_res_spectrum(modelpath, timestep, timestep, angle=angle,
                                                       res_specdata=res_specdata)
                integrated_flux = np.trapz(spectrum['f_lambda'], spectrum['lambda_angstroms'])
                integrated_luminosity = integrated_flux * 4 * np.pi * np.power(u.Mpc.to('cm'), 2)
                bol_luminosity.append(integrated_luminosity)

        lightcurvedata[f'angle={angle}'] = np.log10(bol_luminosity)

    lightcurvedataframe = pd.DataFrame(lightcurvedata)
    lightcurvedataframe = lightcurvedataframe.replace([np.inf, -np.inf], 0)
    print(lightcurvedataframe)

    return lightcurvedataframe


def get_bol_lc_from_lightcurveout(modelpath, res=False):
    if res:
        lcfilename = "light_curve_res.out"
    else:
        lcfilename = "light_curve.out"
    lcdata = pd.read_csv(modelpath / lcfilename, delim_whitespace=True, header=None, names=['time', 'lum', 'lum_cmf'])
    lcdataframes = at.gather_res_data(lcdata, index_of_repeated_value=0)

    times = lcdataframes[0]['time']
    lightcurvedata = {'time': times}

    nangles = len(lcdataframes)
    if not res:
        nangles = 1
    for angle in range(nangles):
        lcdata = lcdataframes[angle]
        bol_luminosity = np.array(lcdata['lum']) * 3.826e33  # Luminosity in erg/s

        # lightcurvedata[f'angle={angle}'] = np.log10(bol_luminosity)
        columnname = 'lum (erg/s)'
        if res:
            columnname = f'angle={angle}'
        lightcurvedata[columnname] = bol_luminosity

    lightcurvedataframe = pd.DataFrame(lightcurvedata)
    lightcurvedataframe = lightcurvedataframe.replace([np.inf, -np.inf], 0)

    return lightcurvedataframe


# modelnames = ['M08_03', 'M08_05', 'M08_10', 'M09_03', 'M09_05', 'M09_10',
#               'M10_02_end55', 'M10_03', 'M10_05', 'M10_10', 'M11_05_1']
modelnames = ['M2a']

for modelname in modelnames:
    # modelpath = Path("/Users/ccollins/harddrive4TB/parameterstudy") / Path(modelname)
    modelpath = Path("/Users/ccollins/harddrive4TB/Gronow2020") / Path(modelname)
    outfilepath = Path("/Users/ccollins/Desktop/bollightcurvedata")

    # lightcurvedataframe = get_bol_lc_from_spec(modelpath)
    lightcurvedataframe = get_bol_lc_from_lightcurveout(modelpath)

    lightcurvedataframe.to_csv(outfilepath / f"bol_lightcurvedata_{modelname}.txt", sep=' ', index=False, header=False)

    with open(outfilepath / f"bol_lightcurvedata_{modelname}.txt", 'r+') as f:  # add comment to start of file
        content = f.read()
        f.seek(0, 0)
        f.write("# 1st col is time in days. Next columns are log10(luminosity) for each model viewing angle".rstrip('\r\n')
                + '\n' + content)

    print("done")



#!/usr/bin/env python3

import argparse
import math
# import os.path

# import numpy as np
import pandas as pd
from astropy import units as u
from pathlib import Path

import artistools as at


def addargs(parser):
    parser.add_argument('-inputpath', '-i',
                        default='.',
                        help='Path of input file')
    parser.add_argument('-outputpath', '-o',
                        default='.',
                        help='Path for output files')


def main(args=None, argsraw=None, **kwargs) -> None:
    if args is None:
        parser = argparse.ArgumentParser(
            formatter_class=at.CustomArgHelpFormatter,
            description='Convert Lapuente model to ARTIS format.')

        addargs(parser)
        parser.set_defaults(**kwargs)
        args = parser.parse_args(argsraw)

    datain_structure = pd.read_csv(Path(args.inputpath, 'structureVMaveraged.txt'),
                                   engine='python', delimiter=r"\s\s+", index_col='i')
    datain_abund = pd.read_csv(Path(args.inputpath, 'abundancesVMaveraged.txt'),
                               delim_whitespace=True)

    datain = datain_structure.join(datain_abund, rsuffix='abund')
    if 'Fe52' not in datain.columns:
        datain.eval('Fe52 = 0.', inplace=True)
    if 'Cr48' not in datain.columns:
        datain.eval('Cr48 = 0.', inplace=True)

    dfmodel = pd.DataFrame(
        columns=['inputcellid', 'velocity_outer', 'logrho', 'X_Fegroup', 'X_Ni56', 'X_Co56',
                 'X_Fe52', 'X_Cr48', 'X_Ni57', 'X_Co57'])
    dfmodel.index.name = 'cellid'
    dfelabundances = pd.DataFrame(columns=['inputcellid', *['X_' + at.get_elsymbol(x) for x in range(1, 31)]])
    dfelabundances.index.name = 'cellid'

    t_model_init_seconds = datain_structure['rad / cm'].iloc[0] / datain_structure['vel / cm/s'].iloc[0]
    t_model_init_days = t_model_init_seconds / 24 / 60 / 60

    tot_ni56mass = 0.
    tot_mass = 0.
    tot_mass2 = 0.
    for cellidin, shell in datain.iterrows():
        cellid = cellidin + 1
        if shell['rho / g/cm^3'] == 0.:
            continue

        v_outer = float(shell['vel / cm/s']) * 1e-5  # convert cm/s to km/s
        rho = shell['rho / g/cm^3']

        tot_mass += shell['shellmass / g']

        tot_ni56mass += shell['shellmass / g'] * shell.Ni56

        abundances = [0. for _ in range(31)]

        for col in datain.columns:
            atomic_number = -1
            if col in at.elsymbols:
                atomic_number = at.elsymbols.index(col)
                abundances[atomic_number] = shell[col]
            elif col.rstrip('0123456789') in at.elsymbols and col.rstrip('0123456789') not in datain.columns:
                # specific nuclide abundance
                atomic_number = at.elsymbols.index(col.rstrip('0123456789'))
                abundances[atomic_number] += shell[col]

        X_fegroup = sum(abundances[26:])
        tot_mass2 += sum(abundances[1:]) * shell['shellmass / g']

        radioabundances = [X_fegroup, shell.Ni56, shell.Co56, shell.Fe52, shell.Cr48, shell.Ni57, shell.Co57]

        dfmodel.loc[cellid] = [cellid, v_outer, math.log10(rho), *radioabundances]
        dfelabundances.loc[cellid] = [cellid, *abundances[1:31]]

    print(f'M_tot  = {tot_mass /  u.solMass.to("g"):.3f} solMass (from sum of specified shell masses)')
    print(f'M_tot  = {tot_mass2 /  u.solMass.to("g"):.3f} solMass (from sum of element densities)')
    print(f'M_Ni56 = {tot_ni56mass /  u.solMass.to("g"):.3f} solMass')

    at.save_modeldata(dfmodel, t_model_init_days, Path(args.outputpath, 'model.txt'))
    at.inputmodel.save_initialabundances(dfelabundances, Path(args.outputpath, 'abundances.txt'))


if __name__ == "__main__":
    main()



from functools import lru_cache
import pandas as pd
import numpy as np
from collections import namedtuple
from pathlib import Path
import artistools


def parse_adata(fadata, phixsdict, ionlist):
    """Generate ions and their level lists from adata.txt."""
    firstlevelnumber = 1

    for line in fadata:
        if not line.strip():
            continue

        ionheader = line.split()
        Z = int(ionheader[0])
        ionstage = int(ionheader[1])
        level_count = int(ionheader[2])
        ionisation_energy_ev = float(ionheader[3])

        if not ionlist or (Z, ionstage) in ionlist:
            level_list = []
            for levelindex in range(level_count):
                row = fadata.readline().split()

                levelname = row[4].strip('\'')
                numberin = int(row[0])
                assert levelindex == numberin - firstlevelnumber
                phixstargetlist, phixstable = phixsdict.get((Z, ionstage, numberin), ([], []))

                level_list.append((float(row[1]), float(row[2]), int(row[3]), levelname, phixstargetlist, phixstable))

            dflevels = pd.DataFrame(level_list,
                                    columns=['energy_ev', 'g', 'transition_count',
                                             'levelname', 'phixstargetlist', 'phixstable'])

            yield Z, ionstage, level_count, ionisation_energy_ev, dflevels

        else:
            for _ in range(level_count):
                fadata.readline()


def parse_transitiondata(ftransitions, ionlist):
    firstlevelnumber = 1

    for line in ftransitions:
        if not line.strip():
            continue

        ionheader = line.split()
        Z = int(ionheader[0])
        ionstage = int(ionheader[1])
        transition_count = int(ionheader[2])

        if not ionlist or (Z, ionstage) in ionlist:
            translist = []
            for _ in range(transition_count):
                row = ftransitions.readline().split()
                translist.append(
                    (int(row[0]) - firstlevelnumber, int(row[1]) - firstlevelnumber,
                     float(row[2]), float(row[3]), int(row[4]) == 1))

            yield Z, ionstage, pd.DataFrame(translist, columns=['lower', 'upper', 'A', 'collstr', 'forbidden'])
        else:
            for _ in range(transition_count):
                ftransitions.readline()


def parse_phixsdata(fphixs, ionlist):
    firstlevelnumber = 1
    nphixspoints = int(fphixs.readline())
    phixsnuincrement = float(fphixs.readline())

    xgrid = np.linspace(1.0, 1.0 + phixsnuincrement * (nphixspoints + 1), num=nphixspoints + 1, endpoint=False)

    for line in fphixs:
        if not line.strip():
            continue

        ionheader = line.split()
        Z = int(ionheader[0])
        upperionstage = int(ionheader[1])
        upperionlevel = int(ionheader[2]) - firstlevelnumber
        lowerionstage = int(ionheader[3])
        lowerionlevel = int(ionheader[4]) - firstlevelnumber
        # threshold_ev = float(ionheader[5])

        assert upperionstage == lowerionstage + 1

        if upperionlevel >= 0:
            targetlist = [(upperionlevel, 1.0)]
        else:
            targetlist = []
            ntargets = int(fphixs.readline())
            for _ in range(ntargets):
                level, fraction = fphixs.readline().split()
                targetlist.append((int(level) - firstlevelnumber, float(fraction)))

        if not ionlist or (Z, lowerionstage) in ionlist:
            phixslist = []
            for _ in range(nphixspoints):
                phixslist.append(float(fphixs.readline()) * 1e-18)
            phixstable = np.array(list(zip(xgrid, phixslist)))

            yield Z, upperionstage, upperionlevel, lowerionstage, lowerionlevel, targetlist, phixstable

        else:
            for _ in range(nphixspoints):
                fphixs.readline()


@lru_cache(maxsize=8)
def get_levels(modelpath, ionlist=None, get_transitions=False, get_photoionisations=False, quiet=False):
    """Return a list of lists of levels."""
    adatafilename = Path(modelpath, 'adata.txt')

    transitionsdict = {}
    if get_transitions:
        transition_filename = Path(modelpath, 'transitiondata.txt')
        if not quiet:
            print(f'Reading {transition_filename.relative_to(modelpath.parent)}')
        with artistools.zopen(transition_filename, 'rt') as ftransitions:
            transitionsdict = {
                (Z, ionstage): dftransitions
                for Z, ionstage, dftransitions in parse_transitiondata(ftransitions, ionlist)}

    phixsdict = {}
    if get_photoionisations:
        phixs_filename = Path(modelpath, 'phixsdata_v2.txt')

        if not quiet:
            print(f'Reading {phixs_filename.relative_to(Path(modelpath).parent)}')
        with artistools.zopen(phixs_filename, 'rt') as fphixs:
            for (Z, upperionstage, upperionlevel, lowerionstage,
                 lowerionlevel, phixstargetlist, phixstable) in parse_phixsdata(fphixs, ionlist):
                phixsdict[(Z, lowerionstage, lowerionlevel)] = (phixstargetlist, phixstable)

    level_lists = []
    iontuple = namedtuple('ion', 'Z ion_stage level_count ion_pot levels transitions')

    with artistools.zopen(adatafilename, 'rt') as fadata:
        if not quiet:
            print(f'Reading {adatafilename.relative_to(Path(modelpath).parent)}')

        for Z, ionstage, level_count, ionisation_energy_ev, dflevels in parse_adata(fadata, phixsdict, ionlist):
            translist = transitionsdict.get((Z, ionstage), pd.DataFrame())
            level_lists.append(iontuple(Z, ionstage, level_count, ionisation_energy_ev, dflevels, translist))

    dfadata = pd.DataFrame(level_lists)
    return dfadata


def parse_recombratefile(frecomb):
    for line in frecomb:
        Z, upper_ionstage, t_count = [int(x) for x in line.split()]
        arr_log10t = []
        arr_rrc_low_n = []
        arr_rrc_total = []
        for _ in range(int(t_count)):
            log10t, rrc_low_n, rrc_total = [float(x) for x in frecomb.readline().split()]

            arr_log10t.append(log10t)
            arr_rrc_low_n.append(rrc_low_n)
            arr_rrc_total.append(rrc_total)

        recombdata_thision = pd.DataFrame({
            'log10T_e': arr_log10t, 'rrc_low_n': arr_rrc_low_n, 'rrc_total': arr_rrc_total})

        recombdata_thision.eval('T_e = 10 ** log10T_e', inplace=True)

        yield Z, upper_ionstage, recombdata_thision


@lru_cache(maxsize=4)
def get_ionrecombratecalibration(modelpath):
    """Read recombrates file."""
    recombdata = {}
    with open(Path(modelpath, 'recombrates.txt'), 'r') as frecomb:
        for Z, upper_ionstage, dfrrc in parse_recombratefile(frecomb):
            recombdata[(Z, upper_ionstage)] = dfrrc

    return recombdata



#!/usr/bin/env python3
# PYTHON_ARGCOMPLETE_OK

import math

import argcomplete
import argparse
from pathlib import Path
import os.path
import pandas as pd
from astropy import units as u
import numpy as np

import artistools as at

MSUN = 1.989e33
CLIGHT = 2.99792458e10


def read_ejectasnapshot(pathtosnapshot):

    column_names = ['id', 'h', 'x', 'y', 'z', 'vx', 'vy', 'vz', 'vstx', 'vsty', 'vstz', 'u',
                    'psi', 'alpha', 'pmass', 'rho', 'p', 'rst', 'tau', 'av', 'ye', 'temp',
                    'prev_rho(i)', 'ynue(i)', 'yanue(i)', 'enuetrap(i)', 'eanuetrap(i)',
                    'enuxtrap(i)', 'iwasequil(i, 1)', 'iwasequil(i, 2)', 'iwasequil(i, 3)']
    ejectasnapshot = pd.read_csv(pathtosnapshot / "ejectasnapshot.dat", delim_whitespace=True, header=None, dtype=float,
                                 names=column_names)
    # Everything is in geometric units here

    return ejectasnapshot


def get_snapshot_time_geomunits(pathtogriddata):
    import glob

    snapshotinfofile = glob.glob(str(Path(pathtogriddata) / "*_info.dat*"))
    if not snapshotinfofile:
        print("No info file found for dumpstep")
        quit()

    if len(snapshotinfofile) > 1:
        print('Too many sfho_info.dat files found')
        quit()
    snapshotinfofile = snapshotinfofile[0]

    if os.path.isfile(snapshotinfofile):
        with open(snapshotinfofile, "r") as fsnapshotinfo:
            line1 = fsnapshotinfo.readline()
            simulation_end_time_geomunits = float(line1.split()[2])
            print(f'Found simulation snapshot time to be {simulation_end_time_geomunits} '
                  f'({simulation_end_time_geomunits * 4.926e-6} s)')

    else:
        print("Could not find snapshot info file to get simulation time")
        quit()

    return simulation_end_time_geomunits


def read_griddat_file(pathtogriddata, targetmodeltime_days=None, minparticlespercell=0):
    griddatfilepath = Path(pathtogriddata) / "grid.dat"

    # Get simulation time for ejecta snapshot
    simulation_end_time_geomunits = get_snapshot_time_geomunits(pathtogriddata)

    griddata = pd.read_csv(griddatfilepath, delim_whitespace=True, comment='#', skiprows=3)
    griddata.rename(columns={
        'gridindex': 'inputcellid',
        'pos_x_min': 'pos_x_min',
        'pos_y_min': 'pos_y_min',
        'pos_z_min': 'pos_z_min',
    }, inplace=True)
    # griddata in geom units
    griddata['rho'] = np.nan_to_num(griddata['rho'], nan=0.)

    if 'cellYe' in griddata:
        griddata['cellYe'] = np.nan_to_num(griddata['cellYe'], nan=0.)
    if 'Q' in griddata:
        griddata['Q'] = np.nan_to_num(griddata['Q'], nan=0.)

    factor_position = 1.478  # in km
    km_to_cm = 1e5
    griddata.eval('pos_x_min = pos_x_min * @factor_position * @km_to_cm', inplace=True)
    griddata.eval('pos_y_min = pos_y_min * @factor_position * @km_to_cm', inplace=True)
    griddata.eval('pos_z_min = pos_z_min * @factor_position * @km_to_cm', inplace=True)

    griddata['rho'] = griddata['rho'] * 6.176e17  # convert to g/cm3

    with open(griddatfilepath, 'r') as gridfile:
        ngrid = int(gridfile.readline().split()[0])
        if ngrid != len(griddata['inputcellid']):
            print("length of file and ngrid don't match")
            quit()
        extratime_geomunits = float(gridfile.readline().split()[0])
        xmax = abs(float(gridfile.readline().split()[0]))
        xmax = (xmax * factor_position) * (u.km).to(u.cm)

    t_model_sec = (simulation_end_time_geomunits + extratime_geomunits) * 4.926e-6  # in seconds
    vmax = xmax / t_model_sec  # cm/s

    t_model_days = t_model_sec / (24. * 3600)  # in days
    print(f"t_model in days {t_model_days} ({t_model_sec} s)")
    corner_vmax = math.sqrt(3 * vmax ** 2)
    print(f"vmax {vmax:.2e} cm/s ({vmax / 29979245800:.2f} * c) per component "
          f"real corner vmax {corner_vmax:.2e} cm/s ({corner_vmax / 29979245800:.2f} * c)")

    if targetmodeltime_days is not None:
        timefactor = targetmodeltime_days / t_model_days
        griddata.eval('pos_x_min = pos_x_min * @timefactor', inplace=True)
        griddata.eval('pos_y_min = pos_y_min * @timefactor', inplace=True)
        griddata.eval('pos_z_min = pos_z_min * @timefactor', inplace=True)
        griddata.eval('rho = rho * @timefactor ** -3', inplace=True)
        print(f"Adjusting t_model to {targetmodeltime_days} days (factor {timefactor}) "
              "using homologous expansion of positions and densities")
        t_model_days = targetmodeltime_days

    if minparticlespercell > 0:
        ncoordgridx = round(len(griddata) ** (1. / 3.))
        xmax = - griddata.pos_x_min.min()
        wid_init = 2 * xmax / ncoordgridx
        filter = np.logical_and(griddata.tracercount < minparticlespercell, griddata.rho > 0.)
        n_ignored = np.count_nonzero(filter)
        mass_ignored = griddata.loc[filter].rho.sum() * wid_init ** 3 / 1.989e33
        mass_orig = griddata.rho.sum() * wid_init ** 3 / 1.989e33

        print(f"Ignoring {n_ignored} nonempty cells ({mass_ignored:.2e} Msun, {mass_ignored / mass_orig * 100:.2f}% of mass) because they have < {minparticlespercell} tracer particles")
        griddata.loc[griddata.tracercount < minparticlespercell, ['rho', 'cellYe']] = 0., 0.

    print(f"Max tracers in a cell {max(griddata['tracercount'])}")

    return griddata, t_model_days, vmax


def mirror_model_in_axis(griddata):
    grid = round(len(griddata) ** (1. / 3.))

    rho = np.zeros((grid, grid, grid))
    cellYe = np.zeros((grid, grid, grid))
    tracercount = np.zeros((grid, grid, grid))
    Q = np.zeros((grid, grid, grid))

    i = 0
    for z in range(0, grid):
        for y in range(0, grid):
            for x in range(0, grid):
                rho[x, y, z] = griddata['rho'][i]
                cellYe[x, y, z] = griddata['cellYe'][i]
                tracercount[x, y, z] = griddata['tracercount'][i]
                Q[x, y, z] = griddata['Q'][i]
                i += 1

    for z in range(0, grid):
        z_mirror = grid-1 - z
        for y in range(0, grid):
            for x in range(0, grid):
                if z < 50:
                    rho[x, y, z] = rho[x, y, z]
                    cellYe[x, y, z] = cellYe[x, y, z]
                    tracercount[x, y, z] = tracercount[x, y, z]
                    Q[x, y, z] = Q[x, y, z]
                if z >= 50:
                    rho[x, y, z] = rho[x, y, z_mirror]
                    cellYe[x, y, z] = cellYe[x, y, z_mirror]
                    tracercount[x, y, z] = tracercount[x, y, z_mirror]
                    Q[x, y, z] = Q[x, y, z_mirror]

    rho_1d_array = np.zeros(len(griddata))
    cellYe_1d_array = np.zeros(len(griddata))
    tracercount_1d_array = np.zeros(len(griddata))
    Q_1d_array = np.zeros(len(griddata))
    i = 0
    for z in range(0, grid):
        for y in range(0, grid):
            for x in range(0, grid):
                rho_1d_array[i] = rho[x, y, z]
                cellYe_1d_array[i] = cellYe[x, y, z]
                tracercount_1d_array[i] = tracercount[x, y, z]
                Q_1d_array[i] = Q[x, y, z]
                i += 1

    griddata['rho'] = rho_1d_array
    griddata['cellYe'] = cellYe_1d_array
    griddata['tracercount'] = tracercount_1d_array
    griddata['Q'] = Q_1d_array

    return griddata


def add_mass_to_center(griddata, t_model_in_days, vmax, args):
    print(griddata)

    # Just (2021) Fig. 16 top left panel
    vel_hole = [0, 0.02, 0.05, 0.07, 0.09, 0.095, 0.1]
    mass_hole = [3e-4, 3e-4, 2e-4, 1e-4, 2e-5, 1e-5, 1e-9]
    mass_intergrated = np.trapz(y=mass_hole, x=vel_hole)  # Msun

    # # Just (2021) Fig. 16 4th down, left panel
    # vel_hole = [0, 0.02, 0.05, 0.1, 0.15, 0.16]
    # mass_hole = [4e-3, 2e-3, 1e-3, 1e-4, 6e-6, 1e-9]
    # mass_intergrated = np.trapz(y=mass_hole, x=vel_hole)  # Msun

    v_outer_hole = 0.1 * CLIGHT  # cm/s
    pos_outer_hole = v_outer_hole * t_model_in_days * (24. * 3600)  # cm
    vol_hole = 4 / 3 * np.pi * pos_outer_hole ** 3  # cm^3
    density_hole = (mass_intergrated * MSUN) / vol_hole  # g / cm^3
    print(density_hole)

    for i, cellid in enumerate(griddata['inputcellid']):
        # if pos < 0.1 c
        if ((np.sqrt(griddata['pos_x_min'][i] ** 2 + griddata['pos_y_min'][i] ** 2 + griddata['pos_z_min'][i] ** 2)) /
                (t_model_in_days * (24. * 3600)) / CLIGHT) < 0.1:
            # if griddata['rho'][i] == 0:
            print("Inner empty cells")
            print(cellid, griddata['pos_x_min'][i], griddata['pos_y_min'][i], griddata['pos_z_min'][i], griddata['rho'][i])
            griddata['rho'][i] += density_hole
            if griddata['cellYe'][i] < 0.4:
                griddata['cellYe'][i] = 0.4
            # print("Inner empty cells filled")
            print(cellid, griddata['pos_x_min'][i], griddata['pos_y_min'][i], griddata['pos_z_min'][i], griddata['rho'][i])

    return griddata


def makemodelfromgriddata(
        gridfolderpath=Path(), outputpath=Path(), targetmodeltime_days=None,
        minparticlespercell=10, getabundances=True, dimensions=3, args=None):

    assert dimensions in [1, 3]
    dfmodel, t_model_days, vmax = at.inputmodel.modelfromhydro.read_griddat_file(
        pathtogriddata=gridfolderpath, targetmodeltime_days=targetmodeltime_days,
        minparticlespercell=minparticlespercell)

    if getattr(args, 'fillcentralhole', False):
        dfmodel = at.inputmodel.modelfromhydro.add_mass_to_center(dfmodel, t_model_days, vmax, args)

    if getattr(args, 'getcellopacityfromYe', False):
        at.inputmodel.opacityinputfile.opacity_by_Ye(outputpath, dfmodel)
    if 'cellYe' in dfmodel:
        at.inputmodel.opacityinputfile.write_Ye_file(outputpath, dfmodel)
    if 'Q' in dfmodel and args.makeenergyinputfiles:
        at.inputmodel.energyinputfiles.write_Q_energy_file(outputpath, dfmodel)

    dfgridcontributions = at.inputmodel.rprocess_from_trajectory.get_gridparticlecontributions(gridfolderpath)

    if getabundances:
        dfmodel, dfelabundances, dfgridcontributions = (
            at.inputmodel.rprocess_from_trajectory.add_abundancecontributions(
                dfgridcontributions=dfgridcontributions, dfmodel=dfmodel, t_model_days=t_model_days,
                minparticlespercell=minparticlespercell))
    else:
        dfelabundances = None

    if dimensions == 1:
        dfmodel, dfelabundances, dfgridcontributions = at.inputmodel.sphericalaverage(
            dfmodel, t_model_days, vmax, dfelabundances, dfgridcontributions)

    if dfgridcontributions is not None:
        at.inputmodel.rprocess_from_trajectory.save_gridparticlecontributions(
            dfgridcontributions, Path(outputpath, 'gridcontributions.txt'))

    if getabundances:
        print('Writing to abundances.txt...')
        at.inputmodel.save_initialabundances(dfelabundances=dfelabundances, abundancefilename=outputpath)
    else:
        at.inputmodel.save_empty_abundance_file(len(dfmodel))

    print('Writing to model.txt...')
    at.inputmodel.save_modeldata(
        modelpath=outputpath, dfmodel=dfmodel, t_model_init_days=t_model_days, dimensions=dimensions, vmax=vmax)


def addargs(parser):
    parser.add_argument('-inputpath', '-i',
                        default='.',
                        help='Path of input files')
    parser.add_argument('-minparticlespercell',
                        default=0, type=int,
                        help='Minimum number of SPH particles in each cell (otherwise set rho=0)')
    parser.add_argument('-dimensions',
                        default=3, type=int,
                        help='Number of dimensions: 1 for spherically symmetric 1D, 3 for 3D Cartesian')
    parser.add_argument('-targetmodeltime_days', '-t', type=float,
                        default=1.,
                        help='Time in days for the output model snapshot')
    parser.add_argument('--noabundances', action='store_true',
                        help='Skip trajectory abundance mapping (get densities only)')
    parser.add_argument('-outputpath', '-o',
                        default=None,
                        help='Path for output model files')


def main(args=None, argsraw=None, **kwargs):
    if args is None:
        parser = argparse.ArgumentParser(
            formatter_class=at.CustomArgHelpFormatter,
            description='Create ARTIS format model from grid.dat.')

        addargs(parser)
        parser.set_defaults(**kwargs)
        argcomplete.autocomplete(parser)
        args = parser.parse_args(argsraw)

    gridfolderpath = args.inputpath
    if not Path(gridfolderpath, 'grid.dat').is_file() or not Path(gridfolderpath, 'gridcontributions.txt').is_file():
        print('grid.dat and gridcontributions.txt are required. Run artistools-maptogrid')
        return
        # at.inputmodel.maptogrid.main()

    if args.outputpath is None:
        outputpath = Path(f'artismodel_{args.dimensions}d')
    else:
        outputpath = Path(args.outputpath)

    outputpath.mkdir(parents=True, exist_ok=True)

    makemodelfromgriddata(
        gridfolderpath=gridfolderpath, outputpath=outputpath, minparticlespercell=args.minparticlespercell,
        targetmodeltime_days=args.targetmodeltime_days, getabundances=(not args.noabundances),
        dimensions=args.dimensions)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# PYTHON_ARGCOMPLETE_OK
from pathlib import Path
import argcomplete
import argparse
import artistools as at
import artistools.inputmodel.downscale3dgrid
import artistools.inputmodel.energyinputfiles
import artistools.inputmodel.modelfromhydro
import artistools.inputmodel.opacityinputfile


def addargs(parser):
    parser.add_argument('-modelpath', default=[], nargs='*', action=at.AppendPath,
                        help='Path to initial model file')

    parser.add_argument('--downscale3dgrid', action='store_true',
                        help='Downscale a 3D ARTIS model to smaller grid size')

    parser.add_argument('-inputgridsize', default=200,
                        help='Size of big model grid for downscale script')

    parser.add_argument('-outputgridsize', default=50,
                        help='Size of small model grid for downscale script')

    parser.add_argument('--makemodelfromgriddata', action='store_true',
                        help='Make ARTIS model files from arepo grid.dat file')

    parser.add_argument('-pathtogriddata', default='.',
                        help='Path to arepo grid.dat file')

    parser.add_argument('--fillcentralhole', action='store_true',
                        help='Fill hole in middle of ejecta from arepo kilonova model')

    parser.add_argument('--getcellopacityfromYe', action='store_true',
                        help='Make opacity.txt where opacity is set in each cell by Ye from arepo model')

    parser.add_argument('--makeenergyinputfiles', action='store_true',
                        help='Downscale a 3D ARTIS model to smaller grid size')

    parser.add_argument('-modeldim', type=int, default=None,
                        help='Choose how many dimensions input model has. 1, 2 or 3')

    parser.add_argument('-outputpath', '-o', default='.',
                        help='Folder for output')


def main(args=None, argsraw=None, **kwargs):
    """Called with makeartismodel. Tools to create an ARTIS input model"""
    if args is None:
        parser = argparse.ArgumentParser(
            formatter_class=at.CustomArgHelpFormatter,
            description='Make ARTIS input model')
        addargs(parser)
        parser.set_defaults(**kwargs)
        argcomplete.autocomplete(parser)
        args = parser.parse_args(argsraw)

    if not args.modelpath:
        args.modelpath = [Path('.')]
    elif isinstance(args.modelpath, (str, Path)):
        args.modelpath = [args.modelpath]

    args.modelpath = at.flatten_list(args.modelpath)

    if args.downscale3dgrid:
        at.inputmodel.downscale3dgrid.make_downscaled_3d_grid(
            modelpath=Path(args.modelpath[0]), inputgridsize=args.inputgridsize, outputgridsize=args.outputgridsize)
        return

    if args.makemodelfromgriddata:
        print(args)
        at.inputmodel.modelfromhydro.makemodelfromgriddata(
            gridfolderpath=args.pathtogriddata, outputpath=args.modelpath[0], getabundances=False, args=args)

    if args.makeenergyinputfiles:
        model, t_model, vmax = at.inputmodel.get_modeldata(args.modelpath[0])
        if args.modeldim == 1:
            rho = 10**model['logrho']
            Mtot_grams = model['cellmass_grams'].sum()

        else:
            rho = model['rho']
            Mtot_grams = model['cellmass_grams'].sum()
        print(f"total mass { Mtot_grams / 1.989e33} Msun")

        at.inputmodel.energyinputfiles.make_energy_files(rho, Mtot_grams, outputpath=args.outputpath)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
import argparse
import multiprocessing
import sys

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from pathlib import Path
# from bigfloat import *
# import numba
# from numpy import arctan as atan

import pynonthermal as pynt

import artistools as at
import artistools.estimators
import artistools.nltepops
import artistools.nonthermal


minionfraction = 0.  # minimum number fraction of the total population to include in SF solution

defaultoutputfile = 'spencerfano_cell{cell:03d}_ts{timestep:02d}_{timedays:.0f}d.pdf'


def make_ntstats_plot(ntstatfile):
    fig, ax = plt.subplots(nrows=1, ncols=1, sharex=True,
                           figsize=(4, 3), tight_layout={"pad": 0.5, "w_pad": 0.3, "h_pad": 0.3})

    dfstats = pd.read_csv(ntstatfile, delim_whitespace=True, escapechar='#')
    dfstats.fillna(0, inplace=True)

    norm_frac_sum = False
    if norm_frac_sum:
        # scale up (or down) ionisation, excitation, and heating to force frac_sum = 1.0
        dfstats.eval("frac_sum = frac_ionization + frac_excitation + frac_heating", inplace=True)
        norm_factors = 1. / dfstats['frac_sum']
    else:
        norm_factors = 1.0
    pd.options.display.max_rows = 50
    pd.options.display.width = 250
    pd.set_option("display.width", 250)
    print(dfstats.to_string())

    xarr = np.log10(dfstats.x_e)
    ax.plot(xarr, dfstats.frac_ionization * norm_factors, label='Ionisation')
    if not max(dfstats.frac_excitation) == 0.:
        ax.plot(xarr, dfstats.frac_excitation * norm_factors, label='Excitation')
    ax.plot(xarr, dfstats.frac_heating * norm_factors, label='Heating')
    ioncols = [col for col in dfstats.columns.values if col.startswith('frac_ionization_')]
    for ioncol in ioncols:
        ion = ioncol.replace('frac_ionization_', '')
        ax.plot(xarr, dfstats[ioncol] * norm_factors, label=f'{ion} ionisation')

    ax.set_ylabel(r'Energy fraction')
    ax.set_xlabel(r'log x$_e$')
    ax.legend(loc='best', handlelength=2, frameon=False, numpoints=1)
    ax.autoscale(enable=True, axis='both', tight=True)
    outputfilename = Path(ntstatfile).with_suffix('.pdf')
    fig.savefig(outputfilename, format='pdf')
    print(f"Saved '{outputfilename}'")
    plt.close()


def addargs(parser):
    parser.add_argument('-modelpath', default='.',
                        help='Path to ARTIS folder')

    parser.add_argument('-timedays', '-time', '-t',
                        help='Time in days to plot')

    parser.add_argument('-timestep', '-ts', type=int,
                        help='Timestep number to plot')

    parser.add_argument('-modelgridindex', '-cell', type=int, default=0,
                        help='Modelgridindex to plot')

    parser.add_argument('-velocity', '-v', type=float, default=-1,
                        help='Specify cell by velocity')

    parser.add_argument('-npts', type=int, default=4096,
                        help='Number of points in the energy grid')

    parser.add_argument('-emin', type=float, default=0.1,
                        help='Minimum energy in eV of Spencer-Fano solution')

    parser.add_argument('-emax', type=float, default=16000,
                        help='Maximum energy in eV of Spencer-Fano solution (approx where energy is injected)')

    parser.add_argument('-vary', action='store', choices=['emin', 'emax', 'npts', 'emax,npts', 'x_e'],
                        help='Which parameter to vary')

    parser.add_argument('-composition', action='store', default='artis', choices=['artis', *at.elsymbols[1:]],
                        help='Composition comes from artis or specific an element to use')

    parser.add_argument('-x_e', type=float, default=2,
                        help='If not using artis composition, specify the electron fraction = N_e / N_ions')

    parser.add_argument('--workfn', action='store_true',
                        help='Testing related to work functions and high energy limits')

    parser.add_argument('--makeplot', action='store_true',
                        help='Save a plot of the non-thermal spectrum')

    parser.add_argument('--differentialform', action='store_true',
                        help=('Solve differential form (KF92 Equation 6) instead of'
                              'integral form (KF92 Equation 7)'))

    parser.add_argument('--noexcitation', action='store_true',
                        help='Do not include collisional excitation transitions')

    parser.add_argument('--atomlossrate', action='store_true',
                        help=('Use Axelrod/Bethe atomic loss rate instead of assuming included cross sections '
                              'are exhaustive'))

    parser.add_argument('--ar1985', action='store_true',
                        help='Use Arnaud & Rothenflug (1985, A&AS, 60, 425) for Fe ionization cross sections')

    parser.add_argument('-o', action='store', dest='outputfile',
                        default=defaultoutputfile,
                        help='Path/filename for PDF file if --makeplot is enabled')

    parser.add_argument('-ostat', action='store',
                        help='Path/filename for stats output')

    parser.add_argument('-plotstats', action='store', default=None,
                        help='Path/filename for NT stats input (no solution, only plotting stat file)')


def main(args=None, argsraw=None, **kwargs):
    if args is None:
        parser = argparse.ArgumentParser(
            formatter_class=at.CustomArgHelpFormatter,
            description='Plot estimated spectra from bound-bound transitions.')
        addargs(parser)
        parser.set_defaults(**kwargs)
        args = parser.parse_args(argsraw)

    if args.plotstats:
        make_ntstats_plot(args.plotstats)
        return

    # global at.nonthermal.experiment_use_Latom_in_spencerfano
    at.nonthermal.experiment_use_Latom_in_spencerfano = args.atomlossrate
    modelpath = Path(args.modelpath)

    if args.workfn:
        return at.nonthermal.workfunction_tests(modelpath, args)

    if Path(args.outputfile).is_dir():
        args.outputfile = Path(args.outputfile, defaultoutputfile)

    if args.composition == 'artis':
        if args.timedays:
            args.timestep = at.get_timestep_of_timedays(modelpath, args.timedays)
        elif args.timestep is None:
            print("A time or timestep must be specified.")
            sys.exit()

        modeldata, _, _ = at.inputmodel.get_modeldata(modelpath)
        if args.velocity >= 0.:
            args.modelgridindex = at.get_mgi_of_velocity_kms(modelpath, args.velocity)
        else:
            args.modelgridindex = args.modelgridindex
        estimators = at.estimators.read_estimators(modelpath, timestep=args.timestep, modelgridindex=args.modelgridindex)
        estim = estimators[(args.timestep, args.modelgridindex)]

        dfpops = at.nltepops.read_files(modelpath, modelgridindex=args.modelgridindex, timestep=args.timestep)

        if dfpops is None or dfpops.empty:
            print(f'ERROR: no NLTE populations for cell {args.modelgridindex} at timestep {args.timestep}')
            return -1

        nntot = estim['populations']['total']
        nne = estim['nne']
        T_e = estim['Te']
        print('WARNING: Use LTE pops at Te for now')
        deposition_density_ev = estim['heating_dep'] / 1.6021772e-12  # convert erg to eV
        ionpopdict = estim['populations']

        velocity = modeldata['velocity_outer'][args.modelgridindex]
        args.timedays = float(at.get_timestep_time(modelpath, args.timestep))
        print(f'timestep {args.timestep} cell {args.modelgridindex} (v={velocity} km/s at {args.timedays:.1f}d)')

    # ionpopdict = {}
    # deposition_density_ev = 327
    # nne = 6.7e5
    #
    # ionpopdict[(26, 1)] = ionpopdict[26] * 1e-4
    # ionpopdict[(26, 2)] = ionpopdict[26] * 0.20
    # ionpopdict[(26, 3)] = ionpopdict[26] * 0.80
    # ionpopdict[(26, 4)] = ionpopdict[26] * 0.
    # ionpopdict[(26, 5)] = ionpopdict[26] * 0.
    # ionpopdict[(27, 2)] = ionpopdict[27] * 0.20
    # ionpopdict[(27, 3)] = ionpopdict[27] * 0.80
    # ionpopdict[(27, 4)] = 0.
    # # ionpopdict[(28, 1)] = ionpopdict[28] * 6e-3
    # ionpopdict[(28, 2)] = ionpopdict[28] * 0.18
    # ionpopdict[(28, 3)] = ionpopdict[28] * 0.82
    # ionpopdict[(28, 4)] = ionpopdict[28] * 0.
    # ionpopdict[(28, 5)] = ionpopdict[28] * 0.

    # x_e = 1.e-2
    # deposition_density_ev = 5.e3
    # nntot = 1.
    # ionpopdict = {}
    # # nne = nntot * x_e
    # # nne = .1
    # dfpops = {}

    # ionpopdict[(at.get_atomic_number('Fe'), 2)] = nntot * 1.
    # ionpopdict[(at.get_atomic_number('Fe'), 3)] = nntot * 0.5
    # ionpopdict[(at.get_atomic_number('Fe'), 4)] = nntot * 0.3

    # KF1992 Figure 2. Pure-Oxygen Plasma
    # x_e = 1.e-2
    # deposition_density_ev = 5.e3
    # nntot = 1.
    # ionpopdict = {}
    # dfpops = {}
    # ionpopdict[(at.get_atomic_number('O'), 1)] = nntot * (1. - x_e)
    # ionpopdict[(at.get_atomic_number('O'), 2)] = nntot * x_e

    # KF1992 Figure 3. Pure-Helium Plasma
    # compelement = args.composition
    # compelement_atomicnumber = at.get_atomic_number(compelement)
    # x_e = args.x_e
    # deposition_density_ev = 5.e3
    # nntot = 1.
    # ionpopdict = {}
    # dfpops = {}
    # ionpopdict[(at.get_atomic_number('He'), 1)] = nntot * (1. - x_e)
    # ionpopdict[(at.get_atomic_number('He'), 2)] = nntot * x_e

    # KF1992 Figure 5. Pure-Iron Plasma
    # x_e = 1.e-2
    # deposition_density_ev = 5.e3
    # nntot = 1.
    # ionpopdict = {}
    # dfpops = {}
    # ionpopdict[(at.get_atomic_number('Fe'), 1)] = nntot * (1. - x_e)
    # ionpopdict[(at.get_atomic_number('Fe'), 2)] = nntot * x_e

    # KF1992 D. The Oxygen-Carbon Zone
    # ionpopdict[(at.get_atomic_number('C'), 1)] = 0.16 * nntot
    # ionpopdict[(at.get_atomic_number('C'), 2)] = 0.16 * nntot * x_e
    # ionpopdict[(at.get_atomic_number('O'), 1)] = 0.82 * nntot
    # ionpopdict[(at.get_atomic_number('O'), 2)] = 0.82 * nntot * x_e
    # ionpopdict[(at.get_atomic_number('Ne'), 1)] = 0.016 * nntot

    # # KF1992 G. The Silicon-Calcium Zone
    # ionpopdict[(at.get_atomic_number('C'), 1)] = 0.38e-5 * nntot
    # ionpopdict[(at.get_atomic_number('O'), 1)] = 0.94e-4 * nntot
    # ionpopdict[(at.get_atomic_number('Si'), 1)] = 0.63 * nntot
    # ionpopdict[(at.get_atomic_number('Si'), 2)] = 0.63 * nntot * x_e
    # ionpopdict[(at.get_atomic_number('S'), 1)] = 0.29 * nntot
    # ionpopdict[(at.get_atomic_number('S'), 2)] = 0.29 * nntot * x_e
    # ionpopdict[(at.get_atomic_number('Ar'), 1)] = 0.041 * nntot
    # ionpopdict[(at.get_atomic_number('Ca'), 1)] = 0.026 * nntot
    # ionpopdict[(at.get_atomic_number('Fe'), 1)] = 0.012 * nntot

    stepcount = 9 if args.vary else 1
    for step in range(stepcount):

        emin = args.emin
        emax = args.emax
        npts = args.npts
        if args.vary == 'emin':
            emin *= 2 ** step
        elif args.vary == 'emax':
            emax *= 2 ** step
        elif args.vary == 'npts':
            npts *= 2 ** step
        elif args.vary == 'x_e':
            assert args.composition != 'artis'
        if args.vary == 'emax,npts':
            npts *= 2 ** step
            emax *= 2 ** step

        if args.composition != 'artis':
            compelement = args.composition
            compelement_atomicnumber = at.get_atomic_number(compelement)
            deposition_density_ev = 5.e3
            nntot = 1.
            x_e = (args.x_e * 10 ** (0.5 * step)) if args.vary == 'x_e' else args.x_e
            ionpopdict = {}
            dfpops = {}
            T_e = 3000
            assert x_e <= 1.
            ionpopdict[(compelement_atomicnumber, 1)] = nntot * (1. - x_e)
            ionpopdict[(compelement_atomicnumber, 2)] = nntot * x_e

        ions = []
        for key in ionpopdict.keys():
            # keep only the ion populations, not element or total populations
            if isinstance(key, tuple) and len(key) == 2 and ionpopdict[key] / nntot >= minionfraction:
                ions.append(key)

        ions.sort()

        if args.noexcitation:
            adata = None
            dfpops = None
        else:
            adata = at.atomic.get_levels(modelpath, get_transitions=True, ionlist=tuple(ions))

        if step == 0 and args.ostat:
            with open(args.ostat, 'w') as fstat:
                strheader = '#emin emax npts x_e frac_sum frac_excitation frac_ionization frac_heating'
                for atomic_number, ion_stage in ions:
                    strheader += ' frac_ionization_' + at.get_ionstring(atomic_number, ion_stage, nospace=True)
                fstat.write(strheader + '\n')

        with pynt.SpencerFanoSolver(
                emin_ev=emin, emax_ev=emax, npts=npts, verbose=True) as sf:

            for Z, ionstage in ions:
                nnion = ionpopdict[(Z, ionstage)]
                if nnion == 0.:
                    print(f'   skipping Z={Z} ion_stage {ionstage} due to nnion={nnion:.1e}')
                    continue

                sf.add_ionisation(Z, ionstage, nnion)
                if not args.noexcitation:
                    sf.add_ion_ltepopexcitation(Z, ionstage, nnion, adata=adata, temperature=T_e)

            sf.solve(depositionratedensity_ev=deposition_density_ev)

            sf.analyse_ntspectrum()

            if args.makeplot:
                # outputfilename = str(args.outputfile).format(
                #     cell=args.modelgridindex, timestep=args.timestep, timedays=args.timedays)
                outputfilename = 'spencerfano.pdf'
                sf.plot_spec_channels(outputfilename=outputfilename)

            if args.ostat:
                with open(args.ostat, 'a') as fstat:
                    strlineout = (f'{emin} {emax} {npts} {x_e:7.2e} {sf.get_frac_sum():6.3f} '
                                  f'{sf.get_frac_excitation_tot():6.3f} {sf.get_frac_ionisation_tot():6.3f} '
                                  f' {sf.get_frac_heating():6.3f}')
                    for atomic_number, ionstage in ions:
                        nnion = ionpopdict[(atomic_number, ionstage)]
                        if nnion > 0.:
                            frac_ionis_ion = sf.get_frac_ionisation_ion(atomic_number, ionstage)
                        else:
                            frac_ionis_ion = 0.
                        strlineout += f' {frac_ionis_ion:.4f}'
                    fstat.write(strlineout + '\n')

    if args.ostat:
        make_ntstats_plot(args.ostat)


if __name__ == "__main__":
    multiprocessing.freeze_support()
    main()



import math

import artistools as at
import numpy as np
import pandas as pd
from pathlib import Path
from astropy import units as u


def get_timestep_times_float(modelpath, loc='mid'):
    modelpath = Path(modelpath)
    _, modelname, codename = modelpath.parts

    filepath = Path(at.config['codecomparisondata1path'], modelname, f"phys_{modelname}_{codename}.txt")

    with open(filepath, "r") as fphys:
        ntimes = int(fphys.readline().replace('#NTIMES:', ''))
        tmids = np.array([float(x) for x in fphys.readline().replace('#TIMES[d]:', '').split()])

    tstarts = np.zeros_like(tmids)
    tstarts[1:] = (tmids[1:] + tmids[:-1]) / 2.
    tstarts[0] = tmids[0] - (tstarts[1] - tmids[0])

    tends = np.zeros_like(tmids)
    tends[:-1] = (tmids[:-1] + tmids[1:]) / 2.
    tends[-1] = tmids[-1] + (tmids[-1] - tstarts[-1])

    if loc == 'mid':
        return tmids
    elif loc == 'start':
        return tstarts
    elif loc == 'end':
        return tends
    elif loc == 'delta':
        tdeltas = tends - tstarts
        return tdeltas
    else:
        raise ValueError("loc must be one of 'mid', 'start', 'end', or 'delta'")


def read_reference_estimators(modelpath, modelgridindex=None, timestep=None):
    """Read estimators from code comparison workshop file.

    """

    virtualfolder, inputmodel, codename = modelpath.parts
    assert virtualfolder == 'codecomparison'

    inputmodelfolder = Path(at.config['codecomparisondata1path'], inputmodel)

    physfilepath = Path(inputmodelfolder, f"phys_{inputmodel}_{codename}.txt")

    estimators = {}
    cell_vel = {}
    cur_timestep = -1
    cur_modelgridindex = -1
    with open(physfilepath, "r") as fphys:
        ntimes = int(fphys.readline().replace('#NTIMES:', ''))
        arr_timedays = np.array([float(x) for x in fphys.readline().replace('#TIMES[d]:', '').split()])
        assert len(arr_timedays) == ntimes

        for line in fphys:
            row = line.split()

            if row[0] == '#TIME:':

                cur_timestep += 1
                cur_modelgridindex = -1
                timedays = float(row[1])
                assert np.isclose(timedays, arr_timedays[cur_timestep], rtol=0.01)

            elif row[0] == '#NVEL:':

                nvel = int(row[1])

            elif not line.lstrip().startswith('#'):

                cur_modelgridindex += 1

                key = (cur_timestep, cur_modelgridindex)
                if key not in estimators:
                    estimators[key] = {'emptycell': False}

                estimators[key]['vel_mid'] = float(row[0])
                estimators[key]['Te'] = float(row[1])
                estimators[key]['rho'] = float(row[2])
                estimators[key]['nne'] = float(row[3])
                estimators[key]['nntot'] = float(row[4])

                estimators[key]['velocity_outer'] = estimators[key]['vel_mid']

    ionfracfilepaths = inputmodelfolder.glob(f"ionfrac_*_{inputmodel}_{codename}.txt")
    for ionfracfilepath in ionfracfilepaths:
        _, element, _, _ = ionfracfilepath.stem.split('_')

        with open(ionfracfilepath, "r") as fions:
            print(ionfracfilepath)
            ntimes_2 = int(fions.readline().replace('#NTIMES:', ''))
            assert ntimes_2 == ntimes

            nstages = int(fions.readline().replace('#NSTAGES:', ''))

            arr_timedays_2 = np.array([float(x) for x in fions.readline().replace('#TIMES[d]:', '').split()])
            assert np.allclose(arr_timedays, arr_timedays_2, rtol=0.01)

            cur_timestep = -1
            cur_modelgridindex = -1
            for line in fions:
                row = line.split()

                if row[0] == '#TIME:':

                    cur_timestep += 1
                    cur_modelgridindex = -1
                    timedays = float(row[1])
                    assert np.isclose(timedays, arr_timedays[cur_timestep], rtol=0.01)

                elif row[0] == '#NVEL:':

                    nvel = int(row[1])

                elif row[0] == '#vel_mid[km/s]':
                    row = [s for s in line.split('  ') if s]  # need a double space because some ion columns have a space
                    iontuples = []
                    ion_startnumber = None
                    for ionstr in row[1:]:
                        atomic_number = at.elsymbols.index(ionstr.strip().rstrip(' 0123456789').title())
                        ion_number = int(ionstr.lstrip('abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ '))

                        # there is unfortunately an inconsistency between codes for
                        # whether the neutral ion is called 0 or 1
                        if ion_startnumber is None:
                            ion_startnumber = ion_number

                        ion_stage = ion_number + 1 if ion_startnumber == 0 else ion_number

                        iontuples.append((atomic_number, ion_stage))

                elif not line.lstrip().startswith('#'):
                    cur_modelgridindex += 1

                    tsmgi = (cur_timestep, cur_modelgridindex)
                    if 'populations' not in estimators[tsmgi]:
                        estimators[tsmgi]['populations'] = {}

                    assert len(row) == nstages + 1
                    assert len(iontuples) == nstages
                    for (atomic_number, ion_stage), strionfrac in zip(iontuples, row[1:]):
                        try:
                            ionfrac = float(strionfrac)
                            ionpop = ionfrac * estimators[tsmgi]['nntot']
                            if ionpop > 1e-80:
                                estimators[tsmgi]['populations'][(atomic_number, ion_stage)] = ionpop
                                estimators[tsmgi]['populations'].setdefault(atomic_number, 0.)
                                estimators[tsmgi]['populations'][atomic_number] += ionpop

                        except ValueError:
                            estimators[tsmgi]['populations'][(atomic_number, ion_stage)] = float('NaN')

                    assert np.isclose(float(row[0]), estimators[tsmgi]['vel_mid'], rtol=0.01)
                    assert estimators[key]['vel_mid']

    return estimators


def get_spectra(modelpath):
    modelpath = Path(modelpath)
    virtualfolder, inputmodel, codename = modelpath.parts
    assert virtualfolder == 'codecomparison'

    inputmodelfolder = Path(at.config['codecomparisondata1path'], inputmodel)

    specfilepath = Path(inputmodelfolder, f"spectra_{inputmodel}_{codename}.txt")

    with open(specfilepath, "r") as fspec:
        ntimes = int(fspec.readline().replace('#NTIMES:', ''))
        nwave = int(fspec.readline().replace('#NWAVE:', ''))
        arr_timedays = np.array([float(x) for x in fspec.readline().split()[1:]])
        assert len(arr_timedays) == ntimes

        dfspectra = pd.read_csv(fspec, delim_whitespace=True, header=None, names=['lambda'] + list(arr_timedays),
                                comment='#')

    return dfspectra, arr_timedays


def plot_spectrum(modelpath, timedays, ax, **plotkwargs):
    dfspectra, arr_timedays = get_spectra(modelpath)
    # print(dfspectra)
    timeindex = (np.abs(arr_timedays - float(timedays))).argmin()
    timedays_found = dfspectra.columns[timeindex + 1]

    print(f"{modelpath}: requested spectrum at {timedays} days. Closest matching spectrum is at {timedays_found} days")
    assert np.isclose(arr_timedays[timeindex], float(timedays_found), rtol=0.01)  # check columns match
    assert np.isclose(float(timedays), float(timedays_found), rtol=0.1)  # found a detect match to requested time
    # print(dfspectra[['lambda', timedays_found]])
    label = str(modelpath).lstrip('_') + f" {timedays_found}d"

    arr_flux = dfspectra[dfspectra.columns[timeindex + 1]] / 4 / math.pi / (u.megaparsec.to('cm') ** 2)

    ax.plot(dfspectra['lambda'], arr_flux, label=label, **plotkwargs)
